#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Submit, get status and history and cancel jobs.

Usage:
    drm4g job submit  [ options ] [ --ntasks <total_tasks> ] [ --dep <job_id> ... ] <template>
    drm4g job list    [ options ] [ <job_id> ]
    drm4g job cancel  [ options ] <job_id> ...
    drm4g job log     [ options ] <job_id>
    drm4g job history [ options ] <job_id>

Arguments:
    <job_id>               Job identifier.
    <template>             Job template.
    <total_tasks>          Total number of tasks in the job array.

Options:
    --ntasks <total_tasks> Number of tasks to submit.
    --dep=<job_id> ...     Define the job dependency list of the job.
    -d --debug             Debug mode.

Commands:
    submit                 Command for submitting jobs.
    list                   Monitor jobs previously submitted.
    cancel                 Cancel jobs.
    log                    Keep track of a job.
    history                Get information about the execution history of a job.

Job field information:
    JID                    Job identification.
    DM 	                   Dispatch Manager state, one of the following: pend, hold, prol, prew, wrap, epil, canl, stop, migr, done, fail.
    EM 	                   Execution Manager state: pend, susp, actv, fail or done.
    START                  The time the job entered the system.
    END                    The time the job reached a final state (fail or done).
    EXEC                   Total execution time, includes suspension time in the remote queue system.
    XFER                   Total file transfer time, includes stage-in and stage-out phases.
    EXIT                   Job exit code.
    TEMPLATE               Filename of the job template used for this job.
    HOST                   Hostname where the job is being executed.
    HID                    Host identification.
    PROLOG                 Total prolog (file stage-in phase) time.
    WRAPPER                Total wrapper (execution phase) time.
    EPILOG                 Total epilog (file stage-out phase) time.
    MIGR                   Total migration time.
    REASON                 The reason why the job left this host.
    QUEUE                  Queue name.
"""

from os.path              import join, exists
from drm4g                import DRM4G_DIR, console_logger, DRM4G_DIR_VAR 
from drm4g.commands       import exec_cmd, Daemon

def run( arg ) :
    try :
        daemon = Daemon( )
        if not daemon.is_alive() :
            raise Exception( 'DRM4G is stopped. ')
        if arg['submit']:
            dependencies = '-d "%s"' % ' '.join( arg['--dep'] ) if arg['--dep'] else ''
            number_of_tasks = '-n %s' % arg['--ntasks'] if arg['--ntasks'] else ''
            cmd = 'gwsubmit %s -v %s %s' % ( dependencies, arg['<template>'], number_of_tasks )
        elif arg['list']:
            cmd = 'gwps -o Jsetxjh '
            if arg['<job_id>'] :
                cmd = cmd + arg['<job_id>'][0]
        elif arg['history']:
            cmd = 'gwhistory %s' % ( arg['<job_id>'][ 0 ] )
        elif arg['log']:
            directory = join(
                              DRM4G_DIR_VAR  ,
                              '%d00-%d99' % ( int(int(float(arg['<job_id>'][0]))/100) , int(int(float(arg['<job_id>'][0]))/100) ) ,
                              arg['<job_id>'][0] ,
                              'job.log'
                            )
            if not exists( directory ) :
                raise Exception( 'There is not a log available for this job.')
            cmd = 'cat %s' % ( directory )
        else :
            cmd = 'gwkill -9 %s' % ( ' '.join( arg['<job_id>'] ) )
        out , err = exec_cmd( cmd )
        console_logger.info( out )
        if err :
            console_logger.info( err )
    except Exception as err :
        console_logger.error( str( err ) )



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import re
import os
import logging
import subprocess
from distutils import spawn


r = re.compile(r'[:,\s]') # match whitespac, coma or :

def parse(output):
    output = [r.split(line) for line in output.splitlines()]
    # now we have a list of lists, but it may contain empty strings
    for line in output:
        while '' in line:
            line.remove('')
    # turn into dict and return
    return dict([(line[0],line[1:]) for line in output])

def exec_cmd( cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE, env = os.environ ):
    logging.debug( "Executing command ... " + cmd )
    p      = subprocess.Popen( cmd, shell = True, stdout = stdout,
                               stderr = stderr, env = env )
    output = p.stdout.read().strip() + p.stderr.read().strip()
    code   = p.wait()
    return code, output

def which( command ):
    """
    Locate commands
    """
    return spawn.find_executable( command )



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Print information about the hosts available on DRM4G.

Usage:
    drm4g [ options ] host [ list ] [ <hid> ]

Arguments:
    <hid>         Host identifier.

Options:
    -d --debug         Debug mode.

Host field information:
    HID           Host identifier.
    ARCH          Architecture.
    JOBS(R/T)     Number of jobs: R = running, T = total.
    LRMS          Local Resource Management System.
    HOSTNAME      Host name.
    QUEUENAME     Queue name.
    WALLT         Queue wall time.
    CPUT          Queue cpu time.
    MAXR          Max. running jobs.
    MAXQ          Max. queued jobs.
"""

from drm4g                import console_logger
from drm4g.commands       import exec_cmd, Daemon

def run( arg ):
    try :
        daemon = Daemon()
        if not daemon.is_alive() :
            raise Exception('DRM4G is stopped.')
        cmd = 'gwhost '
        if arg[ '<hid>' ] :
            cmd = cmd + arg[ '<hid>' ]
        out , err = exec_cmd( cmd )
        console_logger.info( out )
        if err :
            console_logger.info( err )
    except Exception as err :
        console_logger.error( str( err ) )




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import sys
import subprocess
import os
import re
import drm4g.communicators
from drm4g.communicators import ComException
from drm4g.utils.url     import urlparse

import logging
logger  = logging.getLogger(__name__)

import stat
execution_permissions = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH

class Communicator(drm4g.communicators.Communicator):
    """
    Interact with local resources using shell commands
    """
    _lock = __import__('threading').Lock()

    def connect(self):
        logger.debug( "Your are using the local communicator" )

    def execCommand(self, command, input=None ):
        command_proc = subprocess.Popen(command,
            shell = True,
            stdin  = subprocess.PIPE,
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE,
            env = os.environ)
        if input :
            for line in input.split():
                command_proc.stdin.write("%s\n" % line)
                command_proc.stdin.flush()
                stdout, stderr = command_proc.communicate("%s\n" % line)
        else :
            stdout, stderr = command_proc.communicate()
        return stdout.decode() , stderr.decode()

    def mkDirectory(self, url):
        to_dir = self._set_dir(urlparse(url).path)
        out, err = self.execCommand("mkdir -p %s" % to_dir )
        if err:
            output = "Could not create %s directory: %s " % ( to_dir , ' '.join( err.split( '\n' ) ) )
            logger.error( output )
            raise ComException( output )

    def copy(self, source_url, destination_url, execution_mode):
        with self._lock:
            if 'file://' in source_url:
                from_dir = urlparse(source_url).path
                to_dir   = self._set_dir(urlparse(destination_url).path)
            else:
                from_dir = self._set_dir(urlparse(source_url).path)
                to_dir   = urlparse(destination_url).path
            out, err = self.execCommand("cp -r %s %s" % (from_dir,to_dir))
            if err:
                output = "Could not copy from %s to %s : %s" % ( from_dir, to_dir , ' '.join( err.split( '\n' ) ) )
                logger.error( output )
                raise ComException( output )
            if execution_mode == 'X':
                os.chmod(to_dir, execution_permissions )

    def rmDirectory(self, url):
        to_dir   = self._set_dir(urlparse(url).path)
        out, err = self.execCommand("rm -rf %s" % to_dir )
        if err:
            output = "Could not remove %s directory: %s " % ( to_dir , ' '.join( err.split( '\n' ) ) )
            logger.error( output )
            raise ComException( output )

    def checkoutLock(self, url):
        to_dir = self._set_dir(urlparse(url).path)
        return os.path.isfile( '%s/.lock' % to_dir )

    def close(self):
        pass


    #internal
    def _set_dir(self, path):
        work_directory = os.path.expanduser( self.work_directory )
        return re.compile( r'^~' ).sub( work_directory , path )





#!/usr/bin/env python
#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

from drm4g.core.tm_mad import GwTmMad
from optparse import OptionParser
import sys, traceback

def main():
    parser = OptionParser(description = 'Transfer manager MAD',
            prog = 'gw_tm_mad_drm4g.py', version = '0.1',
            usage = 'Usage: %prog')
    options, args = parser.parse_args()
    try:
        GwTmMad().processLine()
    except KeyboardInterrupt as e:
        sys.exit(-1)
    except Exception as e:
        traceback.print_exc(file=sys.stdout)
        exit( 'Caught exception: %s: %s' % (e.__class__, str(e)) )


if __name__ == '__main__':
    main()



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import re
import drm4g.managers
from string import Template


# The programs needed by these utilities. If they are not in a location
# accessible by PATH, specify their location here.
SBATCH  = 'sbatch'   #submit a job
SQUEUE  = 'squeue'   #show status of jobs
SCANCEL = 'scancel'  #delete a job

class Resource (drm4g.managers.Resource):
    pass

class Job (drm4g.managers.Job):

    #job status <--> GridWay job status
    states_SLURM = {'CANCELLED': 'DONE',
                  'COMPLETED' : 'DONE',
                  'COMPLETING': 'ACTIVE',
                  'RUNNING'   : 'ACTIVE',
                  'NODE_FAIL' : 'FAILED',
                  'FAILED'    : 'FAILED',
                  'PENDING'   : 'PENDING',
                  'SUSPENDED' : 'SUSPENDED',
                  'TIMEOUT'   : 'FAILED',
                }

    def jobSubmit(self, pathScript):
        out, err = self.Communicator.execCommand('%s %s' % (SBATCH, pathScript))
        re_job_id = re.compile(r'Submitted batch job (\d*)').search(out)
        if re_job_id:
            return re_job_id.group(1)
        else:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobStatus(self):
        out, err = self.Communicator.execCommand('squeue -h -o %T -j ' + self.JobId)
        if err or not out:
            return 'DONE'
        else:
            return self.states_SLURM.setdefault(out.rstrip('\n'), 'UNKNOWN')

    def jobCancel(self):
        out, err = self.Communicator.execCommand('%s %s' % (SCANCEL, self.JobId))
        if err:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobTemplate(self, parameters):
        args  = '#!/bin/bash\n'
        args += '#SBATCH --job-name=JID_%s\n' % (parameters['environment']['GW_JOB_ID'])
        args += '#SBATCH --output=$stdout\n'
        args += '#SBATCH --error=$stderr\n'
        if parameters['queue'] != 'default':
            args += '#SBATCH -p $queue\n'
        if 'maxWallTime' in parameters :
            args += '#SBATCH --time=%s\n' % (parameters['maxWallTime'])
        if 'maxMemory' in parameters :
            args += '#SBATCH --mem=%s\n' % (parameters['maxMemory'])
        if 'ppn' in parameters :
            args += '#SBATCH --ntasks-per-node=$ppn'
        if 'nodes' in parameters :
            args += '#SBATCH --nodes=$nodes'
        args += '#SBATCH --ntasks=$count\n'
        args += ''.join(['export %s=%s\n' % (k, v) for k, v in list(parameters['environment'].items())])
        args += '\n'
        args += '$executable\n'
        return Template(args).safe_substitute(parameters)





#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import drm4g.managers
from string import Template
import re


# The programs needed by these utilities. If they are not in a location
# accessible by PATH, specify their location here.
MNSUBMIT = 'LANG=POSIX mnsubmit' #mnsubmit - submits a job script to the queue system
MNCANCEL = 'LANG=POSIX mncancel' #mncancel - removes his/her job from the queue system, canceling the execution of the job if it was already running
MNQ      = 'LANG=POSIX mnq'      #mnq      - shows all the jobs submitted

class Resource (drm4g.managers.Resource):
    pass

class Job (drm4g.managers.Job):

    #clock_wall_time is mandatory
    walltime_default = '3600' # 1 hours
    #mn job status <--> GridWay job status
    states_altamira = {
                  'CANCELLED': 'DONE',
                  'COMPLETED' : 'DONE',
                  'COMPLETING': 'ACTIVE',
                  'RUNNING'   : 'ACTIVE',
                  'NODE_FAIL' : 'FAILED',
                  'FAILED'    : 'FAILED',
                  'PENDING'   : 'PENDING',
                  'SUSPENDED' : 'SUSPENDED',
                  'TIMEOUT'   : 'FAILED',
                }

    def jobSubmit(self, pathScript):
        out, err = self.Communicator.execCommand('%s %s' % (MNSUBMIT, pathScript))
        re_job_id = re.compile(r'Submitted batch job (\d*)').search(out)
        if re_job_id:
            return re_job_id.group(1)
        else:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobStatus(self):
        out, err = self.Communicator.execCommand('%s -j %s -h' % (MNQ, self.JobId))
        if err or not out:
            return 'DONE'
        else:
            state = out.split()[3]
            return self.states_altamira.setdefault(state, 'UNKNOWN')

    def jobCancel(self):
        out, err = self.Communicator.execCommand('%s %s' % (MNCANCEL, self.JobId))
        if err:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobTemplate(self, parameters):
        args  = '#!/bin/bash\n'
        args += '# @ job_name = JID_%s\n' % (parameters['environment']['GW_JOB_ID'])
        args += '# @ initialdir = .\n'
        args += '# @ output = $stdout\n'
        args += '# @ error  = $stderr\n'
        args += '# @ total_tasks = $count\n'
        if 'ppn' in parameters :
            args += '# @ tasks_per_node =$ppn\n'
        if 'maxWallTime' in parameters :
            walltime = parameters['maxWallTime']
        else:
            walltime = self.walltime_default
        args += '# @ wall_clock_limit = %s\n' % (walltime)
        args += ''.join(['export %s=%s\n' % (k, v) for k, v in list(parameters['environment'].items())])
        args += '\n'
        args += '$executable\n'
        return Template(args).safe_substitute(parameters)




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import logging


DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

def log_to_file(filename, level=DEBUG):
    l = logging.getLogger("drm4g")
    if len(l.handlers) > 0:
        return
    l.setLevel(level)
    f = open(filename, 'w')
    lh = logging.StreamHandler(f)
    lh.setFormatter(logging.Formatter('%(levelname)-.3s [%(asctime)s.%(msecs)03d] %(name)s: %(message)s',
                                      '%Y%m%d-%H:%M:%S'))
    l.addHandler(lh)

def get_logger(name):
    return logging.getLogger(name)



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import re
import drm4g.managers
from string import Template


# The programs needed by these utilities. If they are not in a location
# accessible by PATH, specify their location here.
BSUB  = 'bsub'   #submit a job
BJOBS = 'bjobs'  #show status of jobs
BKILL = 'bkill'  #delete a job

class Resource (drm4g.managers.Resource):
    pass

class Job (drm4g.managers.Job):

    #job status <--> GridWay job status
    states_LSF = {'DONE'  : 'DONE',
                  'EXIT'  : 'DONE',
                  'RUN'   : 'ACTIVE',
                  'ZOMBI' : 'FAILED',
                  'PEND'  : 'PENDING',
                  'PSUSP' : 'SUSPENDED',
                  'USUSP' : 'SUSPENDED',
                  'SSUSP' : 'SUSPENDED',
                  'UNKWN' : 'UNKNOWN',
                  }

    def jobSubmit(self, pathScript):
        out, err = self.Communicator.execCommand('%s < %s' % (BSUB, pathScript))
        reJobId = re.compile(r'Job <(\d*)> is submitted').search(out)
        if reJobId:
            return reJobId.group(1)
        else:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobStatus(self):
        out, err = self.Communicator.execCommand('%s %s' % (BJOBS, self.JobId))
        if err:
            return 'UNKNOWN'
        else:
            return self.states_LSF.setdefault(out.split()[10], 'UNKNOWN')

    def jobCancel(self):
        out, err = self.Communicator.execCommand('%s %s' % (BKILL, self.JobId))
        if err:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobTemplate(self, parameters):
        args  = '#!/bin/bash\n'
        args += '#BSUB -J JID_%s\n' % (parameters['environment']['GW_JOB_ID'])
        args += '#BSUB -o $stdout\n'
        args += '#BSUB -e $stderr\n'
        args += '#BSUB -n $count\n'
        if parameters['queue'] != 'default':
            args += '#BSUB -q $queue\n'
        args += '#BSUB -W $maxWallTime\n'
        if 'ppn' in parameters:
            args += '#BSUB -R"span[ptile=$ppn]"'
        args += ''.join(['export %s=%s\n' % (k, v) for k, v in list(parameters['environment'].items())])
        args += '\n'
        args += '$executable\n'
        return Template(args).safe_substitute(parameters)



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Start DRM4G's daemon and ssh-agent.

Usage:
    drm4g start [ options ]

Options:
   -d --debug    Debug mode.
"""

from drm4g                import console_logger
from drm4g.commands       import Daemon #, Agent

def run( arg ) :
    try:
        Daemon().start()
        #Agent().start()
    except Exception as err :
        console_logger.error( str( err ) )




# --------------------------------------------------------------------------
# Copyright 2002-2011, GridWay Project Leads (GridWay.org)          
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

from DRMAA import *
import sys, os

#######################
# CREATE JOB TEMPLATE #
#######################

def setup_job_template():

	args=["-l", "-a"]

	(result, jt, error)=drmaa_allocate_job_template()

	if result != DRMAA_ERRNO_SUCCESS:
		print >> sys.stderr, "drmaa_allocate_job_template() failed: %s" % (error)
		sys.exit(-1)

	cwd=os.getcwd()

	(result, error)=drmaa_set_attribute(jt, DRMAA_WD, cwd)
	if result != DRMAA_ERRNO_SUCCESS:
		print >> sys.stderr, "Error setting job template attribute: %s" % (error)
		sys.exit(-1)

	(result, error)=drmaa_set_attribute(jt, DRMAA_JOB_NAME, "ht2")
	(result, error)=drmaa_set_attribute(jt, DRMAA_REMOTE_COMMAND, "/bin/ls")
	(result, error)=drmaa_set_vector_attribute(jt, DRMAA_V_ARGV, args)
  
	if result != DRMAA_ERRNO_SUCCESS:
		print >> sys.stderr,"Error setting remote command arguments: %s" % (error)
		sys.exit(-1)

	(result, error)=drmaa_set_attribute(jt, DRMAA_OUTPUT_PATH, 
					":stdout."+DRMAA_GW_JOB_ID)

	(result, error)=drmaa_set_attribute(jt, DRMAA_ERROR_PATH,
					":stderr."+DRMAA_GW_JOB_ID)

	return jt

#############
# MAIN CODE #
#############

(result, error)=drmaa_init(None)

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_init() failed: %s" % (error)
	sys.exit(-1)
else:
	print "drmaa_init() success"

jt=setup_job_template()

(result, job_id, error)=drmaa_run_job(jt)

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_run_job() failed: %s" % (error)
	sys.exit(-1)

print >> sys.stderr, "Job successfully submited ID: %s" % (job_id)

(result, job_id_out, stat, rusage, error)=drmaa_wait(job_id, DRMAA_TIMEOUT_WAIT_FOREVER)

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_wait() failed: %s" % (error)
	sys.exit(-1)

(result, stat, error)=drmaa_wexitstatus(stat)

print >> sys.stderr, "Job finished with exit code %s, usage: %s" % (stat, job_id)

(result, attr_value)=drmaa_get_next_attr_value(rusage)
while result != DRMAA_ERRNO_NO_MORE_ELEMENTS:
	print >> sys.stderr, "\t%s" % (attr_value)
	(result, attr_value)=drmaa_get_next_attr_value(rusage)

result=drmaa_release_attr_values(rusage)

# ----- Finalize -----

(result, error)=drmaa_delete_job_template(jt)

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_delete_job_template() failed: %s" % (error)
	sys.exit(-1)

(result, error)=drmaa_exit()

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_exit() failed: %s" % (error)
	sys.exit(-1)



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import sys
import logging
from os.path             import join, exists
from drm4g               import DRM4G_DIR, DRM4G_DIR_VAR 
from drm4g.utils.command import exec_cmd


class Job( object ):
    """
    DRM4G job
    """
    args          = dict()
    template_file = None

    def set_name(self, job_name ):
        """
        A name for the job.

        @param job_name: Name of job
        @type  job_name: string
        """
        if not type( job_name ) == str :
            raise Exception( 'Expected a string for job name' )
        else:
            self.args[ 'NAME' ] = job_name

    def get_name(self):
        """
        Get job name
        """
        return self.args.get( 'NAME' )

    def set_executable(self, executable ):
        """
        The executable file

        @param executable: Executable
        @type  executable: string
        """
        if not type( executable ) == str :
            raise Exception( 'Expected a string for executable' )
        else :
            self.args[ 'EXECUTABLE' ] =  executable

    def get_executable(self):
        """
        Get executable file
        """
        return self.args.get( 'EXECUTABLE' )

    def set_arguments(self, arguments ):
        """
        Arguments to the executable

        @param arguments: Executable arguments
        @type  arguments: string
        """
        if not type( arguments ) == str :
            raise Exception( 'Expected a string for arguments' )
        else:
            self.args[ 'ARGUMENTS' ] = arguments

    def get_arguments(self):
        """
        Get executable arguments.
        """
        return self.args.get( 'ARGUMENTS' )

    def set_environment(self, environment):
        """
        User defined environment variables.

        @param environment: Executable arguments
        @type  environment: dictionary
        """
        if not type( environment ) == dict :
            raise Exception( 'Expected a dictionary for environment' )
        else:
            self.args[ 'ENVIRONMENT' ] = environment

    def get_environment(self):
        """
        Get environment variables.
        """
        return self.args.get( 'ENVIRONMENT' )

    def set_input_files(self, files):
        """
        Pair of "local remote" filenames. If the remote filename is missing,
        the local filename will be preserved in the execution host.

        @param files: Input sandbox files
        @type  files: String or list of strings
        """
        if type( files ) == list :
            self.args[ 'INPUT_FILES' ] = files
        elif type( files ) == str :
            self.args[ 'INPUT_FILES' ] = [files]
        else:
            raise Exception( 'Expected a string or a list for input files' )

    def get_input_files(self):
        """
        Get input files.
        """
        return self.args.get( 'INPUT_FILES' )

    def set_output_files(self, files):
        """
        Pair of remote filename local filename. If the local filename is missing,
        the remote filename will be preserved in the client host.

        @param files: Output sandbox files
        @type  files: String or list of strings
        """
        if type( files ) == list :
            self.args[ 'OUTPUT_FILES' ] = files
        elif type( files ) == str :
            self.args[ 'OUTPUT_FILES' ] = [files]
        else:
            raise Exception( 'Expected a string or a list for input files' )

    def get_output_files(self):
        """
        Get output files.
        """
        return self.args.get( 'OUTPUT_FILES' )

    def set_stdin_file(self, stdin_file ):
        """
        Standard input file.

        @param stdin_file: input file
        @type  stdin_file: string
        """
        if not type( stdin_file ) == str :
            raise Exception( 'Expected a string for stdin file' )
        else:
            self.args[ 'STDIN_FILE' ] = stdin_file

    def get_stdin_file(self):
        """
        Get standard input file.
        """
        return self.args.get( 'STDIN_FILE' )

    def set_stdout_file(self, stdout_file ):
        """
        Standard output file.

        @param stdout_file: output file
        @type  stdout_file: string
        """
        if not type( stdout_file ) == str :
            raise Exception( 'Expected a string for stdout file' )
        else:
            self.args[ 'STDOUT_FILE' ] = stdout_file

    def get_stdout_file(self):
        """
        Get standard output file.
        """
        return self.args.get( 'STDOUT_FILE' )

    def set_stderr_file(self, stderr_file ):
        """
        Standard error file.

        @param stderr_file: error file
        @type  stderr_file: string
        """
        if not type( stderr_file ) == str :
            raise Exception( 'Expected a string for error file' )
        else:
            self.args[ 'STDERR_FILE' ] = stderr_file

    def get_stderr_file(self):
        """
        Get standard error file.
        """
        return self.args.get( 'STDERR_FILE' )

    def set_requirements(self, requirements ):
        """
        A boolean expression evaluated for each host, if it
        is true the host will be considered to submit the job. Check the user or
        reference guide to know which variables and operators can be used.

        @param requirements: requirement expression
        @type  requirements: string
        """
        if not type( requirements ) == str :
            raise Exception( 'Expected a string for requirements' )
        else:
            self.args[ 'REQUIREMENTS' ] = requirements

    def get_requirements(self):
        """
        Get requirement expression.
        """
        return self.args.get( 'REQUIREMENTS' )

    def set_np(self, np ):
        """
        Number of process.

        @param np: number of process
        @type  np: string or integer
        """
        if type( np ) == int :
            self.args[ 'NP' ] = str(np)
        elif type( np ) == str :
            self.args[ 'NP' ] = np
        else:
            raise Exception( 'Expected a string or a ingeter for np' )

    def get_np(self):
        """
        Get number of process.
        """
        return self.args.get( 'NP' )

    def set_template_file(self, file ):
        """
        Number of process.

        @param file: number of process
        @type  file: string
        """
        if not type( file ) == str :
            raise Exception( 'Expected a string for fime template' )
        else :
            self.template_file = file

    def get_template_file(self):
        """
        Get template file.
        """
        if not self.template_file :
            return './template.job'
        else :
            return self.template_file

    def create_template(self):
        """
        Create string template.
        """
        str = ''
        for key, val in list(self.args.items()) :
            if key == 'ENVIRONMENT':
                str = str + '%s = %s\n' % ( key, ','.join( "%s %s" % (k,v) for k,v in list(key.items()) ) )
            elif key in ( 'INPUT_FILES', 'OUTPUT_FILES' ):
                str = str + '%s = %s\n' % ( key, ','.join( val ) )
            else :
                str = str + "%s = %s\n" % ( key, val )
        return str

    def create_file(self, string_template ):
        """
        Create file template.
        """
        try :
            f = open( self.get_template_file() , 'w' )
        except :
            raise Exception( "Error creating file template %s" % self.get_template_file() )
        else :
            f.write( string_template )
            f.close()

class DRM4G( object ):
    """
    Class to intact with DRM4G
    """

    def submit(self, job, dep = [] , priority = 0, type_dep = "afterok" ):
        """
        Submit jobs.

        @param job_id: DRM4G job
        @type  job: object
        @param dep: job dependencies
        @type  dep: list
        @param priority: fixed priority
        @type  priority: integer
        @param type_dep: type of dependency : afterok, afterany and afternotok
        @type  type_dep: string
        """
        str_template  = job.create_template()
        job.create_file( str_template )
        depend = "-d %s -r %s" % ( ' '.join( dep ), type_dep ) if dep else ''
        cmd = "gwsubmit -p %d -v %s -t %s" % ( priority,
                                               depend,
                                               job.get_template_file() )
        code, out = exec_cmd( cmd )
        logging.debug( out )
        if code : raise Exception( out )
        job_id = int( out[ 8: ] )
        return job_id

    def get_log(self, job_id ):
        """
        Get log of a job.

        @param job_id: job identifier
        @type  job_id: integer
        """
        job_log = join(DRM4G_DIR_VAR , '%d00-%d99' % ( job_id/100, job_id/100 ), str( job_id ) , 'job.log' )
        if not exists( job_log ) :
            raise Exception( 'There is not a log available for this job.' )
        try :
            f = open( job_log, 'r' )
            return f.readlines()
        finally :
            f.close()

    def status(self, job_id):
        """
        Get job status

        @param job_id: job identifier
        @type  job_id: integer

        JID DM   EM   START    END      EXEC    XFER    EXIT HOST
        33  fail ---- 09:53:18 09:56:57 0:00:00 0:01:00 --   localmachine/fork
        """
        cmd = 'gwps -n -o setxh %d'  % ( job_id )
        code, out = exec_cmd( cmd )
        if code : raise Exception( "Error getting status for job %d" % job_id )
        header = [ 'DM', 'EM', 'START', 'END', 'EXEC', 'XFER', 'EXIT', 'HOST' ]
        return dict( ( key, val ) for key, val in zip( header, out.split() ) )

    def cancel(self, job_id, hard = False):
        """
        Cancel a job.

        @param job_id: job identifier
        @type  job_id: integer
        @param hard: asynchronous cancel option
        @type  hard: boolean
        """
        cmd = "gwkill %s %d" % ( '-9' if hard else '', job_id )
        code, out = exec_cmd( cmd )
        if code : raise Exception( "Error canceling job %d" % job_id )

    def set_priority(self, job_id, priority = 0):
        """
        Set fixed priority to a job.

        @param job_id: job identifier
        @type  job_id: integer
        @param priority: fixed priority
        @type  priority: integer
        """
        cmd = "gwkill -9 -p %d %d" % ( priority, job_id )
        code, out = exec_cmd( cmd )
        if code : raise Exception( "Error updating priority for job %d" % job_id )





# --------------------------------------------------------------------------
# Copyright 2002-2011, GridWay Project Leads (GridWay.org)          
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

from DRMAA import *
import sys, os, time

def print_job_status(id):
	
	# ----- Check job state -----
	
	(result, status, error)=drmaa_job_ps(id)
	
	if result != DRMAA_ERRNO_SUCCESS:
		print >> sys.stderr, "drmaa_job_ps() failed: %s" % (error)
		sys.exit(-1)

	##################################################
	# drmaa_gw_strstatus is not a DRMAA 1.0 function #
	# it is only provided by Gridway                 #
	##################################################

	print "Job state is: " + drmaa_gw_strstatus(status)	

def setup_job_template():
	
	args=["-l", "-a"]
	
	(result, jt, error)=drmaa_allocate_job_template()
	
	if result != DRMAA_ERRNO_SUCCESS:
		print >> sys.stderr, "drmaa_allocate_job_template() failed: %s" % (error)
		sys.exit(-1)

	cwd=os.getcwd()

	# SHOULD CHECK result, REMOVED FOR THE SAKE OF CLARITY

	(result, error)=drmaa_set_attribute(jt, DRMAA_WD, cwd)
	(result, error)=drmaa_set_attribute(jt, DRMAA_JOB_NAME, "ht3")
	(result, error)=drmaa_set_attribute(jt, DRMAA_REMOTE_COMMAND, "/bin/ls")
	(result, error)=drmaa_set_vector_attribute(jt, DRMAA_V_ARGV, args)
	
	(result, error)=drmaa_set_attribute(jt, DRMAA_OUTPUT_PATH, 
					"stdout."+DRMAA_GW_JOB_ID)

	(result, error)=drmaa_set_attribute(jt, DRMAA_ERROR_PATH,
					"stderr."+DRMAA_GW_JOB_ID)

	return jt

########
# MAIN #
########

job_ids=[DRMAA_JOB_IDS_SESSION_ALL]

# ------ INIT DRMAA Library ------
(result, error)=drmaa_init(None)

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_init() failed: %s" % (error)
	sys.exit(-1)

# ------ Set Up job template ------
jt=setup_job_template()

# ------ Set Up job submission state ------
(result, error)=drmaa_set_attribute(jt, 
				DRMAA_JS_STATE, 
				DRMAA_SUBMISSION_STATE_HOLD)

# ------ Submit the job ------
(result, job_id, error)=drmaa_run_job(jt)

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_run_job() failed: %s" % (error)
	sys.exit(-1)

print "Your job has been submitted with id: %s" % (job_id)

# ------ Check job state ------
time.sleep(5)

print_job_status(job_id)

time.sleep(1)

# ------ Control op. ------
print "Releasing the Job"

(result, error)=drmaa_control(job_id, DRMAA_CONTROL_RELEASE)

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_control() failed: %s" % (error)
	sys.exit(-1)

# ------ Check job state again ------
print_job_status(job_id)

# ------ synchronize() ------
print "Synchronizing with job..."

(result, error)=drmaa_synchronize(job_ids, DRMAA_TIMEOUT_WAIT_FOREVER, 0)

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_synchronize failed: %s" % (error)
	sys.exit(-1)

# ------ Control op. --------------------------------------- #
# You can kill the job using dispose=1 in the previous call  #
# You could also use drmaa_wait to get rusage and remove job #
# ---------------------------------------------------------- #

print "Killing the Job"

(result, error)=drmaa_control(job_id, DRMAA_CONTROL_TERMINATE)

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_control() failed: %s" % (error)
	sys.exit(-1)

print "Your job has been deleted"

# ------ Finalize ------
(result, error)=drmaa_delete_job_template(jt)

(result, error)=drmaa_exit()

if result != DRMAA_ERRNO_SUCCESS:
	print >> sys.stderr, "drmaa_exit() failed: %s" % (error)
	sys.exit(-1)



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import drm4g.managers.slurm
from string         import Template


# The programs needed by these utilities. If they are not in a location
# accessible by PATH, specify their location here.
MSUB    = 'msub'   #submit a job
SQUEUE  = 'squeue'   #show status of jobs
SCANCEL = 'scancel'  #delete a job

class Resource (drm4g.managers.slurm.Resource):
    pass

class Job (drm4g.managers.slurm.Job):

    def jobSubmit(self, pathScript):
        out, err = self.Communicator.execCommand('%s %s' % (MSUB, pathScript))
        if out:
            return out.split()[0]
        else:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobTemplate(self, parameters):
        args  = '#!/bin/bash\n'
        args += '#MOAB -N JID_%s\n' % (parameters['environment']['GW_JOB_ID'])
        args += '#MOAB -o $stdout\n'
        args += '#MOAB -e $stderr\n'
        args += '#MOAB -l nodes=$count\n'
        args += '#MOAB -V'
        args += ''.join(['export %s=%s\n' % (k, v) for k, v in list(parameters['environment'].items())])
        args += '\n'
        args += '$executable\n'
        return Template(args).safe_substitute(parameters)





#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import logging
from drm4g import DRM4G_DIR
from os.path import join

REMOTE_JOBS_DIR = join( DRM4G_DIR , 'jobs') #TODO: TO BE DEPRECATED
SSH_PORT = 22                               #TODO: TO BE DEPRECATED
SSH_CONNECT_TIMEOUT = 120 # seconds          TODO: TO BE DEPRECATED
SFTP_CONNECTIONS = 3                        #TODO: TO BE DEPRECATED

logger  = logging.getLogger(__name__)

class ComException(Exception):
    """
    Exception raised by failures in Communicator class
    """
    pass

class Communicator(object):
    """
    Communicator is a abstract class that you must overload for your
    particular communicator. Communicator defines several methods to
    interact with computing resources.
    """
    def __init__(self):
        self.work_directory = REMOTE_JOBS_DIR
        self.port           = SSH_PORT
        self.username       = None
        self.frontend       = None
        self.private_key    = None
        self.public_key     = None

    def connect(self):
        """
        To establish the connection to resource.
        """
        pass

    def execCommand(self, command , input=None ):
        """
        Execute command and return stdout and stderr.

        @param command: a shell command to execute.
        @type command: string
        @param input: optional input argument
        @type input: string
        @return: stdout and stderr associated with the command executed
        @rtype: tuple of string (stdout, stderr)
        """
        pass

    def mkDirectory(self, destination_url):
        """
        Create a directory.

        @param destination_url: url of the folder to create
        @type destination_url: string
        """
        pass

    def copy(self, source_url, destination_url, execution_mode = 'X'):
        """
        Copy a file from source_url to destination_url. If execution_mode = 'X' you
        set execute permission to the destination file.

        @param source_url : file source (url) to copy
        @type source_url: string
        @param destination_url : file destination (url)
        @type destination_url: string
        @param execution_mode : give execute permissions to the file
        @type execution_mode : string
        """
        pass

    def rmDirectory(self, destination_url):
        """
        Remove a directory.

        @param destination_url: url of the folder to remove
        @type destination_url: string
        """
        pass

    def checkoutLock(self, destination_url):
        """
        Check if URL it's locked. 

        @param destination_url: url of the directory/file to check
        @type destination_url: string
        """
        pass


    def close(self):
        """
        Close the connection.
        """
        pass


#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import threading


class List (object):
    """
    Dictionary self-protected.
    """
    def __init__(self):
        self._map = { }
        self._lock = threading.Lock()

    def put(self, key, value):
        self._lock.acquire()
        try:
            self._map[key] = value
        finally:
            self._lock.release()

    def get(self, key):
        self._lock.acquire()
        try:
            return self._map.get(key, None)
        finally:
            self._lock.release()

    def delete(self, key):
        self._lock.acquire()
        try:
            try:
                del self._map[key]
            except KeyError:
                pass
        finally:
            self._lock.release()

    def has_key(self, key):
        self._lock.acquire()
        try:
            return key in self._map
        finally:
            self._lock.release()

    def values(self):
        self._lock.acquire()
        try:
            return list(self._map.values())
        finally:
            self._lock.release()

    def items(self):
        self._lock.acquire()
        try:
            return list(self._map.items())
        finally:
            self._lock.release()







# ------------------------------------------------------------------------------
# Copyright 2002-2011, GridWay Project Leads (GridWay.org)          
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------


import sys
import os
import shutil
import logging
import thread
import time
from arclib import *

logger=None
log_handler=None
job_list={}

# NorduGrid Stuff	

nordugrid_states = {
	'ACCEPTING':	'PENDING',
	'ACCEPTED':		'PENDING',
	'PREPARING':	'PENDING',
	'PREPARED':		'PENDING',
	'SUBMITING':	'PENDING',
	'INLRMS:R':		'ACTIVE',
	'INLRMS:Q':		'ACTIVE',
	'KILLING':		'ACTIVE',
	'EXECUTED FINISHING': 'ACTIVE',
	'FINISHED':		'DONE',
	'FAILED':		'FAILED',
	'KILLED':		'DONE',
	'DELETED':		'DONE'
}

def nordugrid_to_gw_states(state):
	if nordugrid_states.has_key(state):
		return nordugrid_states[state]
	else:
		return 'UNKNOWN'

def get_target_list(rsl, contact, name="fork"):
	# get information about the cluster (with queues)
	ldap_url=URL(contact)
	queues_list=GetQueueInfo(ldap_url)

	# search for the queue
	queue=False
	for i in range(len(queues_list)):
		if queues_list[i].name == name:
			queue=queues_list[i]

	if not queue:
		return None

	# convert the queue into a target and insert it into the target_list	
	target=Target(queue, rsl)
	tl=target_list()
	tl.push_back(target)

	return tl
	
def ldap_contact(host):
	return "ldap://%s:2135/o=grid/mds-vo-name=local" % host

def submit_rsl(rsl, contact, queue_name):
	target_list=get_target_list(rsl, contact, queue_name)
	if target_list==None:
		return None
	job_id=SubmitJob(rsl, target_list)
	return job_id

# Job Class
class GwJob(object):
	def __init__(self, jid):
		self.jid=jid
		self.submited=False
		self.properties_lock=thread.allocate_lock()
		self.downloading=False
		
	def lock(self):
		self.properties_lock.acquire()
		
	def unlock(self):
		self.properties_lock.release()
		
	def submit(self, contact, rsl_file):
		(host, queue) = contact.split("/")
		self.lock()
		self.host=host
		self.queue=queue
		jid=self.jid
		self.unlock()
		
		file_handler=open(rsl_file, 'r')
		rsl_text=file_handler.read()
		file_handler.close()
		rsl=Xrsl(rsl_text)
		
		logger.info("RSL --8<------")
		logger.info(rsl_text)
		logger.info("------>8-- RSL")		
		
		try:
			job_id=submit_rsl(rsl, ldap_contact(host), queue)
			status="PENDING"
			print "SUBMIT %s SUCCESS %s" % (jid, job_id)
			logger.info("Job submited: job_id=%s" % job_id)
			
			self.lock()
			self.job_id=job_id
			self.status=status
			self.unlock()
			
			thread.start_new_thread(self.status_updater, ())
		except Exception, e:
			self.lock()
			self.status="FAILED"
			self.unlock()
			
			print "SUBMIT %s FAILED Error sending job" % jid
			print e
		
		self.lock()
		self.submited=True
		self.unlock()
		
		sys.stdout.flush()
		
	def poll(self):
		if not self.submited:
			print "POLL %s FAILURE Job not yet submited" % self.jid
		else:
			print "POLL %s SUCCESS %s" % (self.jid, self.status)
			
		sys.stdout.flush()
		
	def standalone_poll(self):
		self.lock()
		submited=self.submited
		downloading=self.downloading
		self.unlock()
		
		if not submited:
			print "POLL %s FAILURE Job not yet submited" % self.jid
		elif not downloading:
			self.lock()
			job_id=self.job_id
			self.unlock()
			
			job=GetJobInfo(job_id)
			
			self.lock()
			self.status=nordugrid_to_gw_states(job.status)
			print "POLL %s SUCCESS %s" % (self.jid, self.status)
			self.unlock()
		else:
			print "POLL %s SUCCESS %s" % (self.jid, "ACTIVE")
			
		sys.stdout.flush()
		
	def status_updater(self):
		self.lock()
		submited=self.submited
		job_id=self.job_id
		jid=self.jid
		self.unlock()
		
		if self.submited==True:
			self.lock()
			self.monitor=True
			monitor=self.monitor
			self.unlock()
			
			while monitor:
				
				job=GetJobInfo(job_id)
				new_status=nordugrid_to_gw_states(job.status)
				
				self.lock()
				old_status=self.status
				self.unlock()
				
				if new_status=="DONE" or new_status=="FAILED":
					self.lock()
					self.monitor=False
					self.unlock()
					self.get_output()
					GW_LOCATION=os.environ['GW_LOCATION']
					logger.info(job.exitcode.__class__)
					self.create_wrapper_files(
							job.exitcode, 
							"%s/var/%s/nordugrid" % (GW_LOCATION, jid))
				
				if not new_status==old_status:
					self.lock()
					self.status=new_status
					self.unlock()
					print "CALLBACK %s SUCCESS %s" % (jid, new_status)
					sys.stdout.flush()					
					
				# Change here status polling time
				time.sleep(10)
				
				self.lock()
				monitor=self.monitor
				self.unlock()
				
	def cancel(self):
		self.lock()
		submited=self.submited
		jid=self.jid
		self.unlock()
		if submited==True:
			# Stop it right now or let the monitor die by 
			# itself (after DONE status)?
			self.lock()
			self.monitor=False
			job_id=self.job_id
			self.unlock()
			CancelJob(job_id)
			print "CANCEL %s SUCCESS -" % jid
		else:
			print "CANCEL %s FAILURE Job not submited" % jid
		
		sys.stdout.flush()
		
	def get_output(self):
		self.lock()
		jid=self.jid
		job_id=self.job_id
		self.downloading=True
		self.unlock()
		GW_LOCATION=os.environ['GW_LOCATION']
		destination="%s/var/%s/nordugrid" % (GW_LOCATION, jid)
		logger.info("Downloading dir %s into %s" % (job_id, destination))
		os.mkdir(destination)
		j=JobFTPControl()
		j.DownloadDirectory(URL(job_id), destination)
		
		# copy wrapper stdout & stderr
		# execution -> out
		#shutil.copy(destination+"/stdout.execution", destination+"/stdout.wrapper")
		#shutil.copy(destination+"/stderr.execution", destination+"/stderr.wrapper")
		
		
		
		self.lock()
		self.downloading=False
		self.unlock()
		
	def create_wrapper_files(self, status, directory):
		out_file_name="%s/stdout.wrapper" % directory
		logger.info("Writting: %s" % out_file_name)
		logger.info("Status: %d" % status)
		out_file=open(out_file_name, "w")
		out_file.write("EXIT_STATUS=%d\n" % status)
		out_file.write("REAL_TIME=%d\n" % 10)
		out_file.close()

		err_file_name="%s/stderr.wrapper" % directory
		logger.info("Writting: %s" % err_file_name)
		err_file=open(err_file_name, "w")
		err_file.write("\n\n")
		err_file.close()

		
		
# MAD STUFF

def run(jl=None):
	global logger, log_handler, job_list
	logger=logging.getLogger("em")
	log_handler=logging.FileHandler("/tmp/nordugrid.log")
	log_formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
	log_handler.setFormatter(log_formatter)
	logger.addHandler(log_handler)
	logger.setLevel(logging.INFO)
	
	job_list=jl
	#print job_list
	
	process_line()
	
def stop():
	if logger:
		logger.removeHandler(log_handler)
		log_handler.close()

def process_line():
	line=sys.stdin.readline()
	
	tmp_line=line.split()
	action=""
	jid=""
	contact=""
	rsl_file=""
	
	if len(tmp_line)>0:
		action=tmp_line[0]
		
	if len(tmp_line)>1:
		jid=tmp_line[1]
		
	if len(tmp_line)>2:
		contact=tmp_line[2]
		
	if len(tmp_line)>3:
		rsl_file=" ".join(tmp_line[3:])
	
	process_action(action, jid, contact, rsl_file)
	
def action_init(jid, contact, rsl_file):
	logger.info("INITIALIZATION SUCCESS")
	sys.stdout.write("INIT - SUCCESS -\n")
	sys.stdout.flush()
	
def action_submit(jid, contact, rsl_file):
	global job_list
	if not job_list.has_key(jid):
		job_list[jid]=GwJob(jid)
	logger.info(job_list.__str__())
	
	thread.start_new_thread(job_list[jid].submit, (contact, rsl_file))
	
def action_recover(jid, contact, rsl_file):
	print "RECOVER %s FAILURE Recover not implemented (404)" % jid
	
def action_cancel(jid, contact, rsl_file):
	# print "CANCEL %s FAILURE Cancel not implemented (404)" % jid
	if job_list.has_key(jid):
		thread.start_new_thread(job_list[jid].cancel, ())
	else:
		print "CANCEL %s FAILURE Job not initialized" % jid
		sys.stdout.flush()
	
	
def action_poll(jid, contact, rsl_file):
	# print "POLL %s FAILURE Poll not implemented (404)" % jid
	if job_list.has_key(jid):
		thread.start_new_thread(job_list[jid].standalone_poll, ())
		#thread.start_new_thread(job_list[jid].poll, ())
	else:
		print "POLL %s FAILURE Job not initialized" % jid
		sys.stdout.flush()
		
def action_updater(jid, contact, rsl_file):
	if job_list.has_key(jid):
		thread.start_new_thread(job_list[jid].status_updater, ())
	else:
		print "UPDATER %s FAILURE Job not initialized" % jid
		sys.stdout.flush()
	
		
def action_finalize(jid, contact, rsl_file):
	print "FINALIZE - SUCCESS -"
	sys.exit(0);

action_functions = {
	'INIT': 	action_init,
	'SUBMIT': 	action_submit,
	'RECOVER':	action_recover,
	'CANCEL':	action_cancel,
	'POLL':		action_poll,
	'FINALIZE':	action_finalize,
	'UPDATER':	action_updater
}

def process_action(action, jid, contact, rsl_file):
	logger.info("GOT MESSAGE: \"%s\"" % " ".join([action, jid, contact, rsl_file]))
	logger.info("  Action: %s" % action)
	logger.info("  Jid: %s" % jid)
	logger.info("  Contact: %s" % contact)
	logger.info("  RSL File: %s" % rsl_file)

	tmp_action=action.upper()
	if action_functions.has_key(tmp_action):
		action_functions[tmp_action](jid, contact, rsl_file)
	sys.stdout.flush()



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import sys
import logging
from drm4g.core.configure  import Configuration
from drm4g.utils.message   import Send


class GwImMad (object):
    """
    Information manager MAD

    The format to send a request to the Information MAD, through its standard input, is:

        OPERATION HID HOST ARGS

    Where:
    -OPERATION: Can be one of the following:
        -INIT: Initializes the MAD (i.e. INIT - - -).
        -DISCOVER: Discovers hosts (i.e. DISCOVER - - - ).
        -MONITOR: Monitors a host (i.e. MONITOR HID HOST -).
        -FINALIZE: Finalizes the MAD (i.e. FINALIZE - - -).
    -HID: if the operation is MONITOR, it is a host identifier, chosen by GridWay. Otherwise it is ignored.
    -HOST: If the operation is MONITOR it specifies the host to monitor. Otherwise it is ignored.

    The format to receive a response from the MAD, through its standard output, is:

        OPERATION HID RESULT INFO

    Where:
    -OPERATION: Is the operation specified in the request that originated the response.
    -HID: It is the host identifier, as provided in the submission request.
    -RESULT: It is the result of the operation. Could be SUCCESS or FAILURE.
    -INFO: If RESULT is FAILURE, it contains the cause of failure. Otherwise, if OPERATION
        is   DISCOVER, it contains a list of discovered host, or if OPERATION is MONITOR,
        it contains a list of host attributes.
    """

    logger  = logging.getLogger(__name__)
    message = Send()

    def __init__(self):
        self._resources  = dict()
        self._config     = None

    def do_INIT(self, args):
        """
        Initializes the MAD (i.e. INIT - - -)
        @param args : arguments of operation
        @type args : string
        """
        out = 'INIT - SUCCESS -'
        self.message.stdout(out)
        self.logger.debug(out)

    def do_DISCOVER(self, args, output=True):
        """
        Discovers hosts (i.e. DISCOVER - - -)
        @param args : arguments of operation
        @type args : string
        """
        OPERATION, HID, HOST, ARGS = args.split()
        try:
            self._config  = Configuration()
            self._config.load()
            errors        = self._config.check()
            assert not errors, ' '.join( errors )

            self._resources  = self._config.make_resources()
            communicators    = self._config.make_communicators()
            hosts = ""
            for resname in sorted( self._resources.keys() ) :
                if  self._config.resources[ resname ][ 'enable' ].lower()  == 'false' :
                    continue
                if  'cloud_provider' in self._config.resources[ resname ].keys():
                    continue
                try :
                    self._resources[ resname ][ 'Resource' ].Communicator = communicators[ resname ]
                    self._resources[ resname ][ 'Resource' ].Communicator.connect()
                    hosts = hosts + " " + self._resources[ resname ] [ 'Resource' ].hosts()
                    self._resources[ resname ][ 'Resource' ].Communicator.close()
                except Exception as err :
                    self.logger.error( err , exc_info=1 )
            out = 'DISCOVER %s SUCCESS %s' % ( HID , hosts  )
        except Exception as err :
            out = 'DISCOVER - FAILURE %s' % str( err )
            self.logger.error( err , exc_info=1 )
        if output:
            self.message.stdout( out )
        self.logger.debug( out )

    def do_MONITOR(self, args, output=True):
        """
        Monitors a host (i.e. MONITOR HID HOST -)
        @param args : arguments of operation
        @type args : string
        """
        OPERATION, HID, HOST, ARGS = args.split()
        try:
            info = ""
            for resname, resdict in list(self._resources.items()) :
                if self._config.resources[ resname ][ 'enable' ].lower() == 'false':
                    raise Exception( "Resource '%s' is not enable" % resname )
                if HOST in resdict['Resource'].host_list :
                    info = resdict['Resource'].host_properties( HOST )
                    resdict['Resource'].Communicator.close()
                    break
            assert info, "Host '%s' is not available" % HOST
            out = 'MONITOR %s SUCCESS %s' % (HID , info )
        except Exception as err :
            out = 'MONITOR %s FAILURE %s' % (HID , str(err) )
            self.logger.error( err , exc_info=1 )
        if output:
            self.message.stdout(out)
        self.logger.debug( out )

    def do_FINALIZE(self, args):
        """
        Finalizes the MAD (i.e. FINALIZE - - -)
        @param args : arguments of operation
        @type args : string
        """
        out = 'FINALIZE - SUCCESS -'
        self.message.stdout(out)
        self.logger.debug(out)
        sys.exit(0)

    methods = { 'INIT'    : do_INIT,
                'DISCOVER': do_DISCOVER,
                'MONITOR' : do_MONITOR,
                'FINALIZE': do_FINALIZE,
                }

    def processLine(self):
        """
        Choose the OPERATION through the command line
        """
        try:
            while True:
                input = sys.stdin.readline().split()
                self.logger.debug(' '.join(input))
                OPERATION = input[0].upper()
                if len(input) == 4 and OPERATION in self.methods:
                    self.methods[OPERATION](self, ' '.join(input))
                else:
                    out = 'WRONG COMMAND'
                    self.message.stdout(out)
                    self.logger.debug(out)
        except Exception as err:
            self.logger.warning( str ( err ) , exc_info=1 )




# --------------------------------------------------------------------------
# Copyright 2002-2011, GridWay Project Leads (GridWay.org)          
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

from DRMAA import *
import sys

(result, error)=drmaa_init(None)

if result != DRMAA_ERRNO_SUCCESS:
	print "drmaa_init() failed: "+error
	sys.exit(-1)
else:
	print "drmaa_init() success"
	
	(result, contact, error)=drmaa_get_contact()
	(result, major, minor, error)=drmaa_version()
	(result, drm, error)=drmaa_get_DRM_system()
	(result, impl, error)=drmaa_get_DRMAA_implementation()
  
	print "Using "+impl+", details:"
	print "\t DRMAA version "+str(major)+"."+str(minor)
	print "\t DRMS "+drm+" (contact: "+contact+")"
	
	(result, error)=drmaa_exit()
	
	if result != DRMAA_ERRNO_SUCCESS:
		print "drmaa_exit() failed: "+error
		sys.exit(-1)
	else:
		print "drmaa_exit() success"
	
	sys.exit(0)




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

"""
Manage computing resources on DRM4G.

Usage:
    drm4g resource [ list [ --all ] | edit | check | create | destroy ] [ options ]

 Options:
    -d --debug              Debug mode.
    --all                   Lists all of the created resources.

Commands:
    list                    Show resources available.
    edit                    Configure resouces.
    check                   Check if configured resources are accessible.
    create                  Create new virtual machines
    destroy                 Delete all virtual machines
"""

from drm4g                import console_logger
from drm4g.core.configure import Configuration
from drm4g.commands       import Daemon, Resource

def run( arg ) :
    try :
        config = Configuration()
        resource = Resource( config )
        if arg[ 'edit' ] :
            resource.edit()
        else :
            daemon = Daemon()
            if not daemon.is_alive() :
                raise Exception( 'DRM4G is stopped.' )

            elif arg[ 'check' ] :
                resource.check_frontends( )
            elif arg[ 'create' ] :
                resource.create_vms()
            elif arg[ 'destroy' ] :
                resource.destroy_vms( )
            elif arg[ '--all' ] :
                resource.list_resources( )
            else :
                resource.list()
    except Exception as err :
        console_logger.error( str( err ) )
#TO BE UNCOMMENTED FOR TRACEBACK ERRORS
#        import traceback, sys
#        traceback.print_exc(file=sys.stdout)




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import re
import logging
import drm4g.managers
from os.path         import basename, dirname, join
from drm4g           import REMOTE_VOS_DIR
from drm4g.managers  import JobException, HostInformation, Queue
from time            import sleep 

logger = logging.getLogger(__name__)

X509_USER_PROXY = 'X509_USER_PROXY=' +  join( REMOTE_VOS_DIR , 'x509up.%s ' )
# The programs needed by these utilities. If they are not in a location
# accessible by PATH, specify their location here.
CREAM_DELEGATE = X509_USER_PROXY + 'glite-ce-delegate-proxy'
CREAM_PX_RENEW = X509_USER_PROXY + 'glite-ce-proxy-renew'
CREAM_SUBMIT   = X509_USER_PROXY + 'glite-ce-job-submit'
CREAM_STATUS   = X509_USER_PROXY + 'glite-ce-job-status'
CREAM_DEL      = X509_USER_PROXY + 'glite-ce-job-cancel'
CREAM_PURGE    = X509_USER_PROXY + 'glite-ce-job-purge'
GLOBUS_CP      = X509_USER_PROXY + 'globus-url-copy'

# Regular expressions for parsing.
re_status          = re.compile( "Current Status\s*=\s*\[(.*)\]" )
re_input_files     = re.compile( "GW_INPUT_FILES\s*=\s*\"(.*)\"" )
re_output_files    = re.compile( "GW_OUTPUT_FILES\s*=\s*\"(.*)\"" )
re_executable_file = re.compile( "GW_EXECUTABLE\s*=\s*\"(.*)\"" )
re_obs_url         = re.compile( "CREAM OSB URI\s*=\s*\[(.*)\]" )


def sandbox_files(env_file):

    def parse_files(env, type, re_exp):
        files = []
        match = re_exp.search(env)
        if match :
            for file in match.group( 1 ).split(','):
                if file.startswith( "gsiftp://" ) or file.startswith( "lfn://" ) :
                    continue
                if " " in file:
                    if type == 'output' :
                        file , _  = file.split()
                    else:
                        _  , file = file.split()
                files.append( basename( file ) )
        return files

    with open( env_file , "r" ) as f :
        line_env = ' '.join( f.readlines() )
    input_files      = parse_files( line_env , 'input' ,  re_input_files )
    output_files     = parse_files( line_env , 'output' , re_output_files )
    return input_files, output_files

class Resource (drm4g.managers.Resource):

    def ldapsearch(self, filt = '', attr = '*', bdii = 'lcg-bdii.cern.ch:2170', base = 'Mds-Vo-name=local,o=grid'):
        """
        Wrapper for ldapserch.
        Input parameters:
        filt:    Filter used to search ldap, default = '', means select all
        attr:    Attributes returned by ldapsearch, default = '*', means return all
        host:    Host used for ldapsearch, default = 'lcg-bdii.cern.ch:2170', can be changed by $LCG_GFAL_INFOSYS

        Return each element of list is dictionary with keys:
        'dn':                 Distinguished name of ldapsearch response
        'objectClass':        List of classes in response
        'attr':               Dictionary of attributes
        """
        cmd = 'ldapsearch -x -LLL -l 5 -h %s -b %s "%s" "%s"' % ( bdii, base, filt, attr )
        out, err = self.Communicator.execCommand( cmd )
        assert not err, err

        response = []
        lines = []
        for line in out.split( "\n" ):
            if line.find( " " ) == 0:
                lines[-1] += line.strip()
            else:
                lines.append( line.strip() )
        record = None
        for line in lines:
            if line.find( 'dn:' ) == 0:
                record = {'dn':line.replace( 'dn:', '' ).strip(),
                          'objectClass':[],
                          'attr':{'dn':line.replace( 'dn:', '' ).strip()}}
                response.append( record )
                continue
            if record:
                if line.find( 'objectClass:' ) == 0:
                    record['objectClass'].append( line.replace( 'objectClass:', '' ).strip() )
                    continue
                if line.find( 'Glue' ) == 0:
                    index = line.find( ':' )
                    if index > 0:
                        attr = line[:index]
                        value = line[index + 1:].strip()
                        if attr in record['attr']:
                            if type( record['attr'][attr] ) == type( [] ):
                                record['attr'][attr].append( value )
                            else:
                                record['attr'][attr] = [record['attr'][attr], value]
                        else:
                            record['attr'][attr] = value
        return response

    def hosts(self):
        """
        It will return a string with the host available in the resource.
        """
        if 'vo' in self.features :
            self.host_list = self._hosts_vo( )
            return ' '.join( self.host_list )
        else :
            self.host_list = [ self.name ]
            return self.name

    def _hosts_vo(self):
        """
        It will return a list with the sites available in the VO.
        """
        filt      =   '(&(objectclass=GlueCE)(GlueCEAccessControlBaseRule=VO:%s)(GlueCEImplementationName=CREAM)' % (
                                                                                                       self.features[ 'vo' ] ,
                                                                                                       )
        attr      = 'GlueCEHostingCluster'
        bdii      = self.features.get( 'bdii', '$LCG_GFAL_INFOSYS' )
        result    = []
        if 'host_filter' in self.features :
            for host in self.features[ 'host_filter' ].split( ',' ) :
                ce_filter = '(GlueCEHostingCluster=%s))' % host.strip()
                result.append( self.ldapsearch( filt + ce_filter , attr , bdii ) )
        else :
            result.append( self.ldapsearch( filt + ')' , attr , bdii ) )
        hosts = []
        for value in result :
            for each_host in value :
                host = "%s::%s" % ( self.name , each_host['attr'][attr] )
                hosts.append(host)
        return hosts

    def host_properties(self, host ):
        """
        Obtain the features of each host
        """
        if 'vo' in self.features :
            return self._host_vo_properties( host )
        else :
            return self._host_properties( host )

    def _host_vo_properties(self , host ):
        """
        It will return a string with the features of hosts on Grid environment
        """
        _ , host       = host.split( '::' )
        host_info      = HostInformation()
        host_info.Name = host

        # First search
        filt   =  '(&(objectclass=GlueCE)(GlueCEAccessControlBaseRule=VO:%s)(GlueCEImplementationName=CREAM)(GlueCEInfoHostName=%s))' % ( self.features[ 'vo' ] , host )
        attr   = '*'
        bdii   = self.features.get( 'bdii', '$LCG_GFAL_INFOSYS' )
        result = self.ldapsearch( filt , attr , bdii )

        for value in result :
            queue                = Queue()
            queue.Name           = value['attr']["GlueCEName"]
            queue.Nodes          = value['attr']["GlueCEInfoTotalCPUs"]
            queue.FreeNodes      = value['attr']["GlueCEStateFreeCPUs"]
            queue.MaxTime        = value['attr']["GlueCEPolicyMaxWallClockTime"]
            queue.MaxCpuTime     = value['attr']["GlueCEPolicyMaxCPUTime"]
            queue.MaxJobsInQueue = value['attr']["GlueCEPolicyMaxTotalJobs"]
            queue.MaxRunningJobs = value['attr']["GlueCEPolicyMaxRunningJobs"]
            queue.Status         = value['attr']["GlueCEStateStatus"]
            queue.Priority       = value['attr']["GlueCEPolicyPriority"]
            host_info.addQueue(queue)
            host_info.Nodes      = value['attr']["GlueCEInfoTotalCPUs"]
            host_info.LrmsName   = value['attr']["GlueCEUniqueID"].split('/')[1].rsplit('-',1)[0]
            host_info.LrmsType   = value['attr']["GlueCEInfoLRMSType"]
        host_info.addQueue( Queue() )
        # Second search
        filt   = "(&(objectclass=GlueHostOperatingSystem)(GlueSubClusterName=%s))"  % ( host )
        attr   = '*'
        bdii   = self.features.get( 'bdii', '$LCG_GFAL_INFOSYS' )
        result = self.ldapsearch( filt , attr , bdii )

        try :
            host_info.Os         = result[0]['attr'][ "GlueHostOperatingSystemName" ]
            host_info.OsVersion  = result[0]['attr'][ "GlueHostOperatingSystemVersion" ]
            host_info.Arch       = result[0]['attr'][ "GlueHostArchitecturePlatformType" ]
            host_info.CpuSmp     = result[0]['attr'][ "GlueHostArchitectureSMPSize" ]
        except Exception as err:
            logger.error("The result of '%s' is wrong: %s " % ( filt , str( result ) ) )

        return host_info.info()

class Job (drm4g.managers.Job):

    default_output_files = [
                            'stdout.execution',
                            'stderr.execution',
                            'stdout.wrapper',
                            'stderr.wrapper'
                            ]
    #cream job status <--> GridWay job status
    cream_states = {
                    "REGISTERED"    : "PENDING",
                    "PENDING"       : "PENDING",
                    "IDLE"          : "PENDING",
                    "RUNNING"       : "ACTIVE",
                    "REALLY-RUNNING": "ACTIVE",
                    "HELD"          : "PENDING",
                    "CANCELLED"     : "DONE",
                    "DONE-OK"       : "DONE",
                    "DONE-FAILED"   : "FAILED",
                    "ABORTED"       : "FAILED",
                    }

    def _renew_voms_proxy(self, cont=0):
        output = "The proxy 'x509up.%s' has probably expired" %  self.resfeatures[ 'vo' ]
        logger.debug( output )
        if 'myproxy_server' in self.resfeatures :
            LOCAL_X509_USER_PROXY = "X509_USER_PROXY=%s" % join ( REMOTE_VOS_DIR , self.resfeatures[ 'myproxy_server' ] )
        else :
            LOCAL_X509_USER_PROXY = "X509_USER_PROXY=%s/${MYPROXY_SERVER}" % ( REMOTE_VOS_DIR )
        cmd = "%s voms-proxy-init -ignorewarn -timeout 30 -valid 24:00 -q -voms %s -noregen -out %s" % (
                                                                                                        LOCAL_X509_USER_PROXY ,
                                                                                                        self.resfeatures[ 'vo' ] ,
                                                                                                        join( REMOTE_VOS_DIR , 'x509up.%s ' ) % self.resfeatures[ 'vo' ]
                                                                                                        )
        logger.debug( "Executing command: %s" % cmd )
        out, err = self.Communicator.execCommand( cmd )
        logger.debug( out + err )
        if err :
            output = "Error renewing the proxy(%s): %s" % ( cmd , err )
            logger.error( output )
            if cont<3:
                sleep(2.0)
                self._renew_voms_proxy(cont+1)

    def jobSubmit(self, wrapper_file):
        cmd = '%s -e %s delegete-proxy' % (
                                           CREAM_DELEGATE % self.resfeatures[ 'vo' ] ,
                                           self.resfeatures[ 'host' ]
                                           )
        logger.debug( "Executing command: %s" % cmd )
        out, err = self.Communicator.execCommand( cmd )
        logger.debug( out + err )
        if ( 'The proxy has EXPIRED' in out ) or ( 'is not accessible' in err ) :
            self._renew_voms_proxy()
            logger.debug( "Executing command: %s" % cmd )
            out , err = self.Communicator.execCommand( cmd )
            logger.debug( out + err )
        if ( not 'succesfully delegated' in out ) and ( not 'already exists' in out ) :
            logger.error( out )
            raise JobException( out )
        cmd = '%s -D delegete-proxy -r %s:8443/%s-%s %s' % (
                                     CREAM_SUBMIT % self.resfeatures[ 'vo' ] ,
                                     self.resfeatures[ 'host' ] ,
                                     self.resfeatures[ 'jm' ] ,
                                     self.resfeatures[ 'queue' ] ,
                                     wrapper_file
                                     )
        logger.debug( "Executing command: %s" % cmd )
        out, err = self.Communicator.execCommand( cmd )
        logger.debug( out + err )
        if ( 'The proxy has EXPIRED' in out ) or ( 'is not accessible' in err ) :
            self._renew_voms_proxy()
            logger.debug( "Executing command: %s" % cmd )
            out , err = self.Communicator.execCommand( cmd )
            logger.debug( out + err )
        if not 'https://' in out :
            output = "Error submitting job: %s %s" % ( out, err )
            logger.error( output )
            raise JobException( output )
        return out[ out.find("https://"): ].strip() #cream_id

    def jobStatus(self):
        cmd = '%s -e %s delegete-proxy' % (
                                           CREAM_PX_RENEW % self.resfeatures[ 'vo' ] ,
                                           self.resfeatures[ 'host' ]
                                           )
        logger.debug( "Executing command: %s" % cmd )
        out, err = self.Communicator.execCommand( cmd )
        logger.debug( out + err )
        if not 'succesfully renewed' in out :
            logger.error( out )
            return 'FAILED'
        cmd = '%s %s -L 2' % ( CREAM_STATUS % self.resfeatures[ 'vo' ] , self.JobId )
        logger.debug( "Executing command: %s" % cmd )
        out, err = self.Communicator.execCommand( cmd )
        logger.debug( out + err )
        if 'The proxy has EXPIRED' in out :
            self._renew_voms_proxy()
            logger.debug( "Executing command: %s" % cmd )
            out , err = self.Communicator.execCommand( cmd )
            logger.debug( out + err )
        if "ERROR" in err:
            output = "Error checking '%s' job: %s" % ( self.JobId , err )
            logger.error( output )
            return 'FAILED'
        mo = re_status.search(out)
        if mo:
            job_status = self.cream_states.setdefault(mo.groups()[0], 'UNKNOWN')
            if job_status == 'DONE' or job_status == 'FAILED' :
                output_url = self._getOutputURL( out )
                self._getOutputFiles( output_url )
            return job_status
        else:
            return 'UNKNOWN'

    def jobCancel(self):
        cmd = '%s -N %s' % (CREAM_DEL % self.resfeatures[ 'vo' ]  , self.JobId )
        logger.debug( "Executing command: %s" % cmd )
        out, err = self.Communicator.execCommand( cmd )
        logger.debug( out + err )
        if 'The proxy has EXPIRED' in out :
            self._renew_voms_proxy()
            logger.debug( "Executing command: %s" % cmd )
            out , err = self.Communicator.execCommand( cmd )
            logger.debug( out + err )
        if "ERROR" in err:
            output = "Error canceling '%s' job: %s" % ( self.JobId , err )
            logger.error( output )
            raise JobException( output )

    def jobPurge(self):
        cmd = '%s -N %s' % (CREAM_PURGE % self.resfeatures[ 'vo' ] , self.JobId )
        logger.debug( "Executing command: %s" % cmd )
        out, err = self.Communicator.execCommand( cmd )
        logger.debug( out + err )
        if 'The proxy has EXPIRED' in out :
            self._renew_voms_proxy()
            logger.debug( "Executing command: %s" % cmd )
            out , err = self.Communicator.execCommand( cmd )
        if "ERROR" in err:
            output = "Error purging '%s' job: %s" % ( self.JobId , err )
            logger.error( output )
            raise JobException( output )

    def jobTemplate(self, parameters):
        dir_temp   = self.local_output_directory = dirname( parameters['executable'] )
        executable = basename( parameters['executable'] )
        stdout     = basename( parameters['stdout'] )
        stderr     = basename( parameters['stderr'] )
        count      = parameters['count']
        ppn        = parameters.get( 'ppn', '1' )

        input_sandbox, output_sandbox = sandbox_files( self.resfeatures[ 'env_file' ] )
        if input_sandbox :
            input_files = ' '.join( [ ',"%(dir_temp)s/' + '%s"' % (f) for f in input_sandbox] ) % {'dir_temp':dir_temp }
        else :
            input_files = ''

        self.default_output_files.extend( output_sandbox )
        output_files = ','.join( [ '"%s"' % (f) for f in self.default_output_files ] )

        requirements = ''
        if 'maxWallTime' in parameters:
            requirements += '(other.GlueCEPolicyMaxWallClockTime <= %s)' % parameters['maxWallTime']
        if 'maxCpuTime' in parameters:
            if requirements:
                requirements += ' && '
            requirements += '(other.GlueCEPolicyMaxCPUTime <= %s)' % parameters['maxCpuTime']
        if 'maxMemory' in parameters:
            if requirements:
                requirements += ' && '
            requirements += ' (other.GlueHostMainMemoryRAMSize <= %s)' % parameters['maxMemory']
        Requirements = 'Requirements=%s;' % (requirements) if requirements else ''

        env = ','.join(['"%s=%s"' %(k, v) for k, v in list(parameters['environment'].items())])

        return """
[
JobType = "Normal";
Executable = "%(executable)s";
StdOutput = "%(stdout)s";
StdError = "%(stderr)s";
CpuNumber = %(count)s;
SMPGranularity = %(ppn)s;
OutputSandboxBaseDestURI = "gsiftp://localhost";
InputSandbox = { "%(dir_temp)s/job.env", "%(dir_temp)s/%(executable)s"  %(input_files)s };
OutputSandbox = { %(output_files)s };
Environment = { %(env)s };
%(Requirements)s
]""" % {
        'executable'   : executable,
        'stdout'       : stdout,
        'stderr'       : stderr,
        'dir_temp'     : dir_temp,
        'count'        : count,
        'ppn'          : ppn,
        'input_files'  : input_files,
        'output_files' : output_files,
        'Requirements' : Requirements,
        'env'          : env,
        }

    def _getOutputURL( self, status_output ):
        """
        Resolve the URL for the output files
        """
        match = re_obs_url.search( status_output )
        if match :
            url = match.group( 1 )
            return url
        else :
            output = "Output URL not found in '%s'" % ( status_output )
            logger.error( output )
            raise JobException( output )

    def _getOutputFiles( self, output_url ):
        """
        Get output files from the remote output_url
        """
        for file in self.default_output_files :
            cmd = '%s %s file://%s' % (
                                       GLOBUS_CP % self.resfeatures[ 'vo' ],
                                       join( output_url , file ) ,
                                       join( self.local_output_directory , file )
                                       )
            logger.debug( "Coping file '%s' : %s" % ( file , cmd ) )
            out, err = self.Communicator.execCommand( cmd )
            if 'error' in err :
                output = "Error coping file '%s' : %s" % ( file , err )
                logger.error( output )
                raise JobException( output )





#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import xml.dom.minidom
import re
import drm4g.managers
from string import Template


# The programs needed by these utilities. If they are not in a location
# accessible by PATH, specify their location here.
MNSUBMIT = 'LANG=POSIX mnsubmit' #mnsubmit - submits a job script to the queue system
CHECKJOB = 'LANG=POSIX checkjob' #checkjob - obtains detailed information about a specific job
MNCANCEL = 'LANG=POSIX mncancel' #mncancel - removes his/her job from the queue system, canceling the execution of the job if it was already running
MNQ      = 'LANG=POSIX mnq'      #mnq      - shows all the jobs submitted

class Resource (drm4g.managers.Resource):
    pass

class Job (drm4g.managers.Job):

    #clock_wall_time is mandatory
    walltime_default = '172800' # 48 hours
    #mn job status <--> GridWay job status
    states_mn = {
                  'BatchHold' : 'PENDING',
                  'SystemHold': 'PENDING',
                  'UserHold'  : 'PENDING',   #Job is idle and is not eligible to run due to a user, admin, or batch system hold.
                  'Canceling' : 'PENDING',   #Job is in the process of being cancelled.
                  'Removed'   : 'FAILED',    #Job has run to its requested walltime successfully but has been canceled by the scheduler or resource manager due to exceeding its walltime or violating another policy; includes jobs canceled by users or administrators either before or after a job has started.
                  'NotQueued' : 'FAILED',    #Indicates a system problem in most cases.
                  'Vacated'   :	'FAILED',    #Job canceled after walltime=partial execution due to a system failure.
                  'Running'   : 'ACTIVE',    #Job is currently executing the user application
                  'Migrated'  : 'PENDING',   #This is a transitional state that indicates that the job is in being handed off to the native SLURM resource manager on a specific machine in preparation for running
                  'Idle'      : 'PENDING',   #Job is queued and eligible to run but is not yet executing
                  'Deferred'  : 'FAILED',    #Job can not be run for one reason or another, however Moab will will continue to evaluate the job periodically for run eligibility.
                  'Staging'   :	'PENDING',   #The job has been submitted by Moab to the native batch system for it to run and the native batch system has not confirmed yet that the job is actually running.
                  'Starting'  : 'PENDING',   #The batch system has attempted to start the job and the job is currently performing pre-start tasks which may including provisioning resources, staging data,executing system pre-launch scripts, etc.
                  'Completed' :	'DONE',      #Job has completed running
                  'Suspended' : 'SUSPENDED', #Job was running but has been suspended by the scheduler or an admin.
                }

    def jobSubmit(self, pathScript):
        out, err = self.Communicator.execCommand('%s %s' % (MNSUBMIT, pathScript))
        re_job_id = re.compile(r'Submitted batch job (\d*)').search(err)
        if re_job_id:
            return re_job_id.group(1)
        else:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobStatus(self):
        out, err = self.Communicator.execCommand('%s %s --xml' % (CHECKJOB, self.JobId))
        if err:
            return 'DONE'
        else:
            out_parser = xml.dom.minidom.parseString(out)
            state = out_parser.getElementsByTagName('Data')[0].getElementsByTagName('job')[0].getAttribute('EState')
            return self.states_mn.setdefault(state, 'UNKNOWN')

    def jobCancel(self):
        out, err = self.Communicator.execCommand('%s %s' % (MNCANCEL, self.JobId))
        if not "job '%s' cancelled" % (self.JobId) in out:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobTemplate(self, parameters):
        args  = '#!/bin/bash\n'
        args += '# @ job_name = JID_%s\n' % (parameters['environment']['GW_JOB_ID'])
        args += '# @ initialdir = $directory\n'
        args += '# @ output = $stdout\n'
        args += '# @ error  = $stderr\n'
        args += '# @ total_tasks = $count\n'
        if 'ppn' in parameters :
            args += '# @ tasks_per_node =$ppn\n'
        if 'maxWallTime' in parameters :
            walltime = parameters['maxWallTime']
        else:
            walltime = self.walltime_default
        args += '# @ wall_clock_limit = %s\n' % (walltime)
        args += ''.join(['export %s=%s\n' % (k, v) for k, v in list(parameters['environment'].items())])
        args += '\n'
        args += '$executable\n'
        return Template(args).safe_substitute(parameters)




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import time
import pickle
import logging
import threading
import drm4g.managers
import drm4g.managers.fork
from utils                  import load_json
from os.path                import exists, join
from drm4g                  import DRM4G_DIR, DRM4G_LOGGER_CONF, RESOURCE_MANAGERS
from drm4g.utils.importlib  import import_module
try:
    from configparser       import SafeConfigParser
except ImportError:
    from ConfigParser       import SafeConfigParser  # ver. < 3.0

logger = logging.getLogger(__name__)

pickled_file = join(DRM4G_DIR, "var", "rocci_pickled")

lock = threading.RLock()

def pickle_remove(inst, resource_name):
    with lock:
        try:
            instances=[]
            with open( pickled_file+"_"+resource_name, "r" ) as pf :
                while True :
                    try:
                        instances.append( pickle.load( pf ) )
                    except EOFError :
                        break
            if not instances :
                logger.error( "There are no VMs defined in '%s' or the file is not well formed." % (pickled_file+"_"+resource_name) )
                exit( 1 )

            with open( pickled_file+"_"+resource_name, "w" ) as pf :
                for instance in instances:
                    if instance.ext_ip != inst.ext_ip :
                        pickle.dump( instance, pf )

            if len(instances) == 1 :
                os.remove( pickled_file+"_"+resource_name )
        except Exception as err:
            logger.error( "Error deleting instance from pickled file %s\n%s" % (pickled_file+"_"+resource_name, str( err )) )

def start_instance( instance, resource_name ) :
    with lock:
        try:
            instance.create()
            instance.get_ip()
            with open( pickled_file+"_"+resource_name, "a" ) as pf :
                pickle.dump( instance, pf )
        except Exception as err :
            logger.error( "Error creating instance: %s" % str( err ) )
            try :
                logger.debug( "Trying to destroy the instance" )
                instance.delete( )
            except Exception as err :
                logger.error( "Error destroying instance\n%s" % str( err ) )

def stop_instance( instance, resource_name ):
    try :
        instance.delete()
        pickle_remove(instance, resource_name)
    except Exception as err :
        logger.error( "Error destroying instance\n%s" % str( err ) )

def manage_instances(args, resource_name, config):
    if args == "start" :
        try :
            hdpackage = import_module( RESOURCE_MANAGERS[config['lrms']] + ".%s" % config['lrms'] )
        except Exception as err :
            raise Exception( "The infrastructure selected does not exist. "  + str( err ) )
        threads = []
        handlers = []
        try:
            instance = eval( "hdpackage.Instance( config )" )
        except KeyError as err:
            logger.error( "Either you have defined an incorrect value in your configuration file 'resources.conf'" \
                " or there's a value that doesn't correspond with any of the keys in your cloud setup file 'cloudsetup.json':" )
            raise
        except Exception as err:
            logger.error( "An error ocurred while trying to create a VM instance." )
            raise
        for number_of_th in range( int(config['instances']) ):
            th = threading.Thread( target = start_instance, args = ( instance, resource_name, ) )
            th.start()
            threads.append( th )
        [ th.join() for th in threads ]
    elif args == "stop" :
        instances = []
        if not exists( pickled_file+"_"+resource_name ):
            logger.error( "There are no available VMs to be deleted for the resource %s" % resource_name )
        else:
            with open( pickled_file+"_"+resource_name, "r" ) as pf :
                while True :
                    try:
                        instances.append( pickle.load( pf ) )
                    except EOFError :
                        break
            if not instances :
                logger.error( "There are no VMs defined in '%s' or the file is not well formed." % (pickled_file+"_"+resource_name) )
                exit( 1 )
            threads = []
            for instance in instances :
                th = threading.Thread( target = stop_instance, args = ( instance, resource_name ) )
                th.start()
                threads.append( th )
            [ th.join() for th in threads ]
    else :
        logger.error( "Invalid option" )
        exit( 1 )

class Resource (drm4g.managers.Resource):
    def hosts(self):
        """
        It will return a string with the host available in the resource.
        """
        if 'cloud_provider' in self.features :
            self.host_list = [ "" ]
            return ""
        else :
            self.host_list = [ self.name ]
            return self.name

class Job (drm4g.managers.fork.Job):
    pass

class CloudSetup(object):

    def __init__(self, name, features = {}):
        self.name             = name
        self.vo               = features.get( "vo" )
        #self.url              = features.get( "url" )
        self.cloud_providers  = features.get( "cloud_providers" )



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

try :
    import Queue as queue
except :
    import queue
import sys
import traceback
from threading        import Thread
from threading        import Lock


"""
Use example
from dynamic import ThreadPool
pool = ThreadPool(2, 10)
pool.add_task(func, arg)
"""

threads_active = 0
threads_min    = 1
timeout        = 1
lock           = Lock()

class Worker(Thread):
    """
    Thread executing tasks from a given tasks queue
    """
    def __init__(self, tasks):
        Thread.__init__(self)
        self.tasks = tasks
        self.setDaemon(True)
        self.start()

    def run(self):
        while True:
            try:
                try:
                    func, args, kargs = self.tasks.get( timeout  )
                except queue.Empty:
                    with lock :
                        global threads_active, threads_min
                        if threads_active > threads_min:
                            threads_active -= 1
                            break
                        else :
                            continue
                except Exception:
                    sys.exit(-1)
                else:
                    try:
                        func(*args, **kargs)
                    except :
                        traceback.print_exc(file=sys.stdout)
            except Exception:
                pass

class ThreadPool:
    """
    Pool of threads consuming tasks from a queue
    """
    def __init__(self, num_threads_min, num_threads_max):
        self.threads_max = num_threads_max
        self.tasks = queue.Queue()
        with lock :
            global threads_active, threads_min
            threads_active =  num_threads_min
            threads_min    = num_threads_min
        for _ in range(num_threads_min):
            Worker(self.tasks)

    def add_task(self, func, *args, **kargs):
        """
        Add a task to the queue
        """
        with lock :
            global threads_active
            try:
                if threads_active < self.threads_max:
                    Worker(self.tasks)
                    threads_active += 1
                self.tasks.put((func, args, kargs))
            except :
                traceback.print_exc(file=sys.stdout)




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import sys
import xml.dom.minidom
import os
import subprocess
import pickle


import logging
logger  = logging.getLogger(__name__)


def totalCores( cores ):
    return sum( [ int( core ) for core in cores.split( ',' ) ] )

def sec_to_H_M_S( sec ):
    """
    Convert seconds into HH:MM:SS
    """
    m, s = divmod( int( sec ), 60)
    h, m = divmod( m, 60 )
    return "%d:%02d:%02d" % ( h, m, s )

class ResourceException(Exception):
    pass

class JobException(Exception):
    pass

class Resource (object):
    """
    Class to obtain information about compute resources
    """

    def __init__(self):
        self.name           = None
        self.features       = dict()
        self.Communicator   = None
        self.host_list      = []

    def hosts(self):
        """
        It will return a string with the host available in the resource.
        """
        self.host_list = [ self.name ]
        return self.name

    def host_properties(self, host ):
        """
        Obtain the features of each host
        """
        return self._host_properties( host )

    def _host_properties(self , host ):
        """
        It will return a string with the features of hosts on HPC environment
        """
        host_info       = HostInformation()
        host_info.Name  = host
        host_info.Name, host_info.OsVersion, host_info.Arch, host_info.Os = self.system_information()

        q_features = [ ( q_elem.strip(), jobr_elem.strip(), jobq_elem.strip() )
                      for q_elem, jobr_elem, jobq_elem in zip(
                                                  self.features[ 'queue' ].split( ',' ) ,
                                                  self.features[ 'max_jobs_running' ].split( ',' ) ,
                                                  self.features[ 'max_jobs_in_queue' ].split( ',' )
                                                  ) ]
        for queue_name, max_jobs_running, max_jobs_in_queue in q_features  :
            queue                = Queue()
            queue.Name           = queue_name
            queue.MaxRunningJobs = max_jobs_running
            queue.MaxJobsInQueue = max_jobs_in_queue
            host_info.addQueue( self.additional_queue_properties( queue ) )
 #       host_info.addQueue( Queue() )
        host_info.LrmsName = self.features[ 'lrms' ]
        host_info.LrmsType = self.features[ 'lrms' ]
        return host_info.info()

    def system_information(self):
        """
        It will return a tuple with hostname, OS version, architecture, OS name
        """
        out, err = self.Communicator.execCommand('uname -n -r -m -o')
        if not err:
            return out.split()
        else:
            logger.error("Error executing `uname` command: %s" % ' '.join( err.split( '\n' ) ) )
            return ('NULL', 'NULL', 'NULL', 'NULL')

    # To overload
    def additional_queue_properties(self, Queue):
        return Queue

class Job (object):
    """
    Class to manage jobs
    """

    def __init__(self):
        self.resfeatures  = dict()
        self.Communicator = None
        self.JobId        = None
        self._status      = None
        self._lock        = __import__('threading').Lock()

    def setStatus(self, status):
        with self._lock :
            self._status = status

    def getStatus(self):
        with self._lock :
            return self._status

    def refreshJobStatus(self):
        with self._lock :
            self._status = self.jobStatus()

    def get_abs_directory(self, directory ):
        out, err = self.Communicator.execCommand( 'ls -d %s' % directory )
        if not err:
            return out.strip('\n')
        else:
            output = "Could not obtain  the '%s' directory : %s" % ( directory , str ( err ) )
            logger.error( output )
            raise JobException( output )

    def createWrapper(self, local_directory, template):
        try:
            f = open(local_directory, 'w')
        except Exception as e:
            raise JobException('Error creating wrapper_drm4g :' + str(e))
        else:
            f.write(template)
            f.close()

    def copyWrapper(self, local_directory, remote_directory):
        """
        Copy wrapper_drm4g file to the remote host
        """
        source_url      = 'file://%s'       % local_directory
        destination_url = 'gsiftp://_/%s' % remote_directory
        try:
            self.Communicator.copy(source_url, destination_url, 'X')
        except Exception as e:
            raise JobException("Error copying wrapper_drm4g : %s" % str(e) )

    # To overload
    def jobSubmit( self , wrapper_dir ) :
        pass

    def jobStatus( self ) :
        pass

    def jobCancel( self ) :
        pass

    def jobTemplate( self , rsl ) :
        pass

class Queue( object ) :

    def __init__(self):
        self.Name           = ""
        self.Nodes          = "0"
        self.FreeNodes      = "0"
        self.MaxTime        = "0"
        self.MaxCpuTime     = "0"
        self.MaxCount       = "0"
        self.MaxRunningJobs = "0"
        self.MaxJobsInQueue = "0"
        self.Status         = "NULL"
        self.DispatchType   = "NULL"
        self.Priority       = "NULL"

    def info (self, i):
        """
        @return: the information of the queue
        @rtype: string
        """
        i = str(i)
        return  'QUEUE_NAME[' + i + ']="' + self.Name + '" QUEUE_NODECOUNT[' + i + ']=' + self.Nodes + \
            ' QUEUE_FREENODECOUNT[' + i + ']=' + self.FreeNodes + ' QUEUE_MAXTIME[' + i +']=' + self.MaxTime + \
            ' QUEUE_MAXCPUTIME[' + i + ']=' + self.MaxCpuTime + ' QUEUE_MAXCOUNT[' + i + ']=' + self.MaxCount + \
            ' QUEUE_MAXRUNNINGJOBS[' + i + ']=' + self.MaxRunningJobs + ' QUEUE_MAXJOBSINQUEUE[' + i + ']=' + \
            self.MaxJobsInQueue  + ' QUEUE_STATUS[' + i + ']="' + self.Status + '" QUEUE_DISPATCHTYPE[' + i + ']="'+ \
            self.DispatchType + '" QUEUE_PRIORITY[' + i +']="' + self.Priority + '" '

class HostInformation( object ) :

    def __init__(self):
        self.Name       = ""
        self.Arch       = "NULL"
        self.Os         = "NULL"
        self.OsVersion  = "NULL"
        self.CpuModel   = "NULL"
        self.CpuMhz     = "0"
        self.FreeCpu    = "0"
        self.CpuSmp     = "0"
        self.Nodes      = "0"
        self.SizeMemMB  = "0"
        self.FreeMemMB  = "0"
        self.SizeDiskMB = "0"
        self.FreeDiskMB = "0"
        self.ForkName   = "NULL"
        self.LrmsName   = "NULL"
        self.LrmsType   = "NULL"
        self._queues    = []

    def addQueue(self, queue):
        self._queues.append(queue)

    def showQueues(self):
        return self._queues

    def info(self):
        """
        @return: the information of the host and the host queues
        @rtype: string
        """
        info_host = 'HOSTNAME="' + self.Name + '" ARCH="' + self.Arch + '" OS_NAME="' + \
            self.Os + '" OS_VERSION="' + self.OsVersion + '" CPU_MODEL="' + self.CpuModel + \
            '" CPU_MHZ=' + self.CpuMhz + ' CPU_FREE=' + self.FreeCpu + ' CPU_SMP=' + self.CpuSmp + \
            ' NODECOUNT=' + self.Nodes + ' SIZE_MEM_MB=' + self.SizeMemMB + ' FREE_MEM_MB=' + \
            self.FreeMemMB + ' SIZE_DISK_MB=' + self.SizeDiskMB + ' FREE_DISK_MB=' + self.FreeDiskMB + \
            ' FORK_NAME="' + self.ForkName + '" LRMS_NAME="' + self.LrmsName + '" LRMS_TYPE="' + self.LrmsType + '" FIXED_PRIORITY="0" '
        info_queue = ""
        for i, queue in enumerate(self._queues):
            info_queue += queue.info(i)
        return info_host + info_queue




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import sys
import logging
import threading
from drm4g.utils.url       import urlparse
from drm4g.utils.dynamic   import ThreadPool
from drm4g.core.configure  import Configuration
from drm4g.utils.message   import Send


class GwTmMad (object):
    """
    Transfer manager MAD

    The File Transfer Driver interfaces with Grid Data Management Services
    and is responsible for file staging, remote working directory set-up
    and remote host clean up.

    The format to send a request to the Transfer MAD, through its standard
    input, is:

    OPERATION JID TID EXE_MODE SRC_URL DST_URL

    Where:

    -OPERATION: Can be one of the following:
        -INIT: Initializes the MAD, JID should be max number of jobs.
        -START: Init transfer associated with job JID
        -END: Finish transfer associated with job JID
        -MKDIR: Creates directory SRC_URL
        -RMDIR: Removes directory SRC_URL
        -CP: start a copy of SRC_URL  to DST_URL, with identification TID,
            and associated with job JID.
        -FINALIZE: Finalizes the MAD.
    -JID: Is a job identifier, chosen by GridWay.
    -TID: Transfer identifier, only relevant for command CP.
    -EXE_MODE: If equal to 'X' file will be given execution permissions,
        only relevant for command CP.

    The format to receive a response from the MAD, through its standard
    output, is:

    OPERATION JID TID RESULT INFO

    Where:

    -OPERATION: Is the operation specified in the request that originated
        the response or CALLBACK, in the case of an asynchronous notification
        of a state change.
    -JID: It is the job identifier, as provided in the START request.
    -TID: It is the transfer identifier, as provided in the CP request.
    -RESULT: It is the result of the operation. Could be SUCCESS or FAILURE.
    -INFO: If RESULT is FAILURE, it contains the cause of failure.

    """

    logger  = logging.getLogger(__name__)
    message = Send()

    def __init__(self):
        self._max_thread   = 30
        self._min_thread   = 5
        self._lock         = threading.Lock()
        self._communicator = dict()
        self._configure    = None

    def do_INIT(self, args):
        """
        INIT: Initializes the MAD, JID should be max number of jobs.
        (i.e. INIT JID - - - -)
        @param args : arguments of operation
        @type args : string
        """
        out = 'INIT - - SUCCESS -'
        self.message.stdout( out )
        self.logger.debug( out )

    def do_START(self, args):
        """
        START: Init transfer associated with job JID.(i.e. START JID - - - -)
        @param args : arguments of operation
        @type args : string
        """
        out = 'START %s - SUCCESS -' % ( args.split( ) [ 1 ] )
        self.message.stdout( out )
        self.logger.debug( out )

    def do_END(self, args):
        """
        END: Finish transfer associated with job JID .(i.e. END JID - - - -)
        @param args : arguments of operation
        @type args : string
        """
        out = 'END %s - SUCCESS -' % ( args.split( ) [ 1 ] )
        self.message.stdout( out )
        self.logger.debug( out )

    def do_FINALIZE(self, args):
        """
        Finalizes the MAD (i.e. FINALIZE - - - - -)
        @param args : arguments of operation
        @type args : string
        """
        out = 'FINALIZE %s - SUCCESS -' % ( args.split( ) [ 1 ] )
        self.message.stdout( out )
        self.logger.debug( out )
        sys.exit( 0 )

    def do_MKDIR(self, args):
        """
        MKDIR: Creates directory SRC_URL (i.e. MKDIR JID - - SRC_URL -)
        @param args : arguments of operation
        @type args : string
        """
        OPERATION, JID, TID, EXE_MODE, SRC_URL, DST_URL = args.split()
        try:
            com = self._update_com( urlparse( SRC_URL ).host )
            com.rmDirectory( SRC_URL )
            com.mkDirectory( SRC_URL )
            out = 'MKDIR %s - SUCCESS -' % ( JID )
        except Exception as err :
            out = 'MKDIR %s - FAILURE %s' % ( JID , str( err ) )
            self.logger.error( err , exc_info=1 )
        self.message.stdout( out )
        self.logger.debug( out )

    def do_RMDIR(self, args):
        """
        RMDIR: Removes directory SRC_URL (i.e. RMDIR JID - - SRC_URL -)
        @param args : arguments of operation
        @type args : string
        """
        OPERATION, JID, TID, EXE_MODE, SRC_URL, DST_URL = args.split()
        try:
            com = self._update_com( urlparse( SRC_URL ).host )
            if not com.checkoutLock(SRC_URL):
                com.rmDirectory(SRC_URL)
                out = 'RMDIR %s - SUCCESS -' % (JID)
            else:
                out = 'RMDIR %s - FAILURE Directory "%s" is locked' % (JID, SRC_URL)
        except Exception as err :
            out = 'RMDIR %s - FAILURE %s' % ( JID , str( err ) )
            self.logger.error( err , exc_info=1 )
        self.message.stdout( out )
        self.logger.debug( out )

    def do_CP(self, args):
        """
        CP: start a copy of SRC_URL  to DST_URL, with identification TID,
        and associated with job JID.(i.e. CP JID TID - SRC_URL DST_URL)
        @param args : arguments of operation
        @type args : string
        """
        OPERATION, JID, TID, EXE_MODE, SRC_URL, DST_URL = args.split()
        if 'file:' in SRC_URL:
            url = DST_URL
        else:
            url = SRC_URL
        try:
            com = self._update_com( urlparse( url ).host )
            com.copy( SRC_URL , DST_URL , EXE_MODE )
            out = 'CP %s %s SUCCESS -' % ( JID , TID )
        except Exception as err :
            out = 'CP %s %s FAILURE %s' % ( JID , TID , str( err ) )
            self.logger.error( err , exc_info=1 )
        self.message.stdout( out )
        self.logger.debug(out)

    methods = {'INIT'    : do_INIT,
               'START'   : do_START,
               'END'     : do_END,
               'MKDIR'   : do_MKDIR,
               'RMDIR'   : do_RMDIR,
               'CP'      : do_CP,
               'FINALIZE': do_FINALIZE}

    def processLine(self):
        """
        Choose the OPERATION through the command line
        """
        try:
            pool = ThreadPool( self._min_thread , self._max_thread )
            self._configure = Configuration()
            while True:
                input = sys.stdin.readline().split()
                self.logger.debug(' '.join(input))
                OPERATION = input[0].upper()
                if len(input) == 6 and OPERATION in self.methods:
                    if OPERATION == 'FINALIZE' or OPERATION == 'INIT':
                        self.methods[OPERATION](self, ' '.join(input))
                    else: pool.add_task(self.methods[OPERATION], self,' '.join(input))
                else:
                    out = 'WRONG COMMAND'
                    self.message.stdout(out)
                    self.logger.debug(out)
        except Exception as err :
            self.logger.warning( str ( err ) , exc_info=1 )

    def _update_com(self, host):
        with self._lock :
            self.logger.debug( "_update_com function - It's looking for a communicator for the host: %s" % host )
            if not self._configure:
                self._configure = Configuration()
            '''
            if self._configure.check_update() or not self._configure.resources :
                self._configure.load()
                errors = self._configure.check()
                if errors :
                    self.logger.error ( ' '.join( errors ) )
                    raise Exception ( ' '.join( errors ) )
            '''
            #This needs to be optimized
            self._configure.load()
            errors = self._configure.check()
            if errors :
                self.logger.error ( ' '.join( errors ) )
                raise Exception ( ' '.join( errors ) )
            for resname, resdict in list(self._configure.resources.items()) :
                self.logger.debug( "    The current resource to which it's being compared is %s" % resname )
                if 'cloud_provider' in self._configure.resources[ resname ].keys():
                    continue
                elif '::' in host :
                    _resname , _ = host.split( '::' )
                    if resname != _resname :
                        continue
                elif resname != host :
                    continue
                elif resname not in self._communicator:
                    self.logger.debug( "    Since they are the same, its Communicator() will be returned")
                    self._communicator[ resname ] = self._configure.make_communicators()[resname]
                self.logger.debug( "\nCommunicator for %s:\n    communicator: %s\n    username: %s\n    frontend: %s" % (resname, resdict[ 'communicator' ], self._communicator[ resname ].username, self._communicator[ resname ].frontend) )
                return self._communicator[ resname ]



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import sys
from os.path     import dirname, abspath, join, expanduser, exists
import fabric

import re
import socket
import drm4g.commands
import drm4g.communicators
from drm4g.communicators    import ComException, SFTP_CONNECTIONS
from drm4g.utils.url        import urlparse

import logging
logger  = logging.getLogger(__name__)

class Communicator(drm4g.communicators.Communicator):
    """
    Create a SSH session to remote resources.
    """
    _lock = __import__('threading').Lock()
    _sem = __import__('threading').Semaphore(SFTP_CONNECTIONS)
    _conn = None

    def connect(self):
        with self._lock:
            if not self.is_connected:
                self._connect()

    def _connect(self):
        logger.debug("Opening ssh connection ... ")
        logger.debug("Trying '%s' key ... " % self.private_key )
        private_key_path = expanduser(self.private_key)
        if (
                (not exists( private_key_path)) and
                ('PRIVATE KEY' not in self.private_key)
        ):
            output = "'%s'key does not exist" % private_key_path
            raise ComException(output)
        logger.debug(
            "Connecting to '%s' as user '%s' port  '%s' ..."
            % ( self.frontend, self.username, self.port )
        )
        if ':' in self.frontend:
            self.frontend, self.port = self.frontend.split( ':' )
        try:
            self._conn = fabric.Connection(
                host=self.frontend,
                user=self.username,
                connect_kwargs={
                    "key_filename": private_key_path,
                },
                connect_timeout=300
            )
        except Exception as err:
            logger.warning(
                "Error connecting '%s': %s" %
                (self.frontend, str(err))
            )

    @property
    def is_connected(self):
        if self._conn is None:
            return False
        else:
            return self._conn.is_connected

    def execCommand(self, command, input=None):
        self.connect()
        with self._lock:
            result = self._conn.run(command)
            stdout = result.stdout
            stderr = result.stderr
        return stdout, stderr

    def mkDirectory(self, url):
        to_dir = self._set_dir(urlparse(url).path)
        stdout, stderr = self.execCommand("mkdir -p %s" % to_dir)
        if stderr:
            raise ComException("Could not create %s directory on '%s': %s" % (to_dir, self.frontend , stderr ))

    def rmDirectory(self, url):
        to_dir = self._set_dir(urlparse(url).path)
        stdout, stderr = self.execCommand("rm -rf %s" % to_dir)
        if stderr:
            raise ComException("Could not remove %s directory on '%s': %s" % (to_dir, self.frontend , stderr ))

    def copy(self, source_url, destination_url, execution_mode=''):
        with self._sem:
            self.connect()
            if 'file://' in source_url:
                from_dir = urlparse(source_url).path
                to_dir = self._set_dir(urlparse(destination_url).path)
                with self._lock:
                    self._conn.put(from_dir, to_dir)
                if execution_mode == 'X':
                    stdout, stderr = self.execCommand("chmod +x %s" % to_dir)
            else:
                from_dir = self._set_dir(urlparse(source_url).path)
                to_dir = urlparse(destination_url).path
                logger.warning("%s , %s" % (from_dir, to_dir))
                with self._lock:
                    self._conn.get(from_dir, to_dir)

    def close(self):
        try:
            if self.is_connected:
                self._conn.close()
        except Exception as err:
            logger.warning("Could not close the SSH connection to '%s': %s" % (self.frontend, str(err)))

    def __del__(self):
        """
        Attempt to clean up if not explicitly closed.
        """
        self.close()

    #internal
    def _set_dir(self, path):
        work_directory = re.compile(r'^~').sub(self.work_directory, path)
        if work_directory[0] == r'~':
            return ".%s" % work_directory[1:]
        else:
            return work_directory



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import re
import drm4g.managers
from string import Template


# The programs needed by these utilities. If they are not in a location
# accessible by PATH, specify their location here.
QCONF = 'LANG=POSIX qconf'
QHOST = 'LANG=POSIX qhost'
QSUB  = 'LANG=POSIX qsub'
QSTAT = 'LANG=POSIX qstat'
QDEL  = 'LANG=POSIX qdel'

class Resource (drm4g.managers.Resource):

    def additional_queue_properties(self, queue):
        out, err = self.Communicator.execCommand('%s -sq %s' % (QCONF, queue.Name))
        if err:
            raise drm4g.managers.ResourceException(' '.join(err.split('\n')))
        reWalltime = re.compile(r'h_rt\s*(\d+):(\d+):\d+')
        if reWalltime.search(out):
            try:
                hours, minutes   = reWalltime.search(out).groups()
                queue.MaxTime    = str(int(hours) * 60 + int(minutes))
            except:
                pass
        reCputime = re.compile(r'h_cpu\s*(\d+):(\d+):\d+')
        if reCputime.search(out):
            try:
                hours, minutes   = reCputime.search(out).groups()
                queue.MaxCpuTime = str(int(hours) * 60 + int(minutes))
            except:
                pass
        return queue

class Job (drm4g.managers.Job):

    #sge job status <--> GridWay job status
    states_sge = {
        'd'  : 'FAILED',   #An error occurred with the job
        'Eqw': 'FAILED',   #An error occurred with the job
        'Ew' : 'FAILED',   #An error occurred with the job
        'h'  : 'PENDING',  #Job is hold
        'r'  : 'ACTIVE',   #Job is running
        'R'  : 'FAILED',   #Job is restarted
        'Rr' : 'FAILED',   #Job is restarted
        's'  : 'SUSPENDED',#Job is suspended
        'S'  : 'SUSPENDED',#Job is suspended
        't'  : 'PENDING',  #Job is transfering
        'T'  : 'PENDING',  #Job is Threshold
        'w'  : 'PENDING',  #Job is waiting
        'qw' : 'PENDING',  #Job is waiting
        }

    def jobSubmit(self, pathScript):
        out, err = self.Communicator.execCommand('%s %s' % (QSUB, pathScript))
        if err:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))
        re_job  = re.compile(r'^Your job (\d*) .*').search(out)
        return re_job.group(1)

    def jobStatus(self):
        out, err = self.Communicator.execCommand('%s | grep %s' % (QSTAT, self.JobId))
        if not out:
            return 'DONE'
        else:
            state = out.split()[4]
            return self.states_sge.setdefault(state, 'UNKNOWN')

    def jobCancel(self):
        out, err = self.Communicator.execCommand('%s %s' % (QDEL, self.JobId))
        if err:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobTemplate(self, parameters):
        args  = '#!/bin/bash\n'
        args += '#$ -N JID_%s\n' % (parameters['environment']['GW_JOB_ID'])
        if 'project' in parameters :
            args += '#$ -P $project\n'
        if parameters['queue'] != 'default':
            args += '#$ -q $queue\n'
        args += '#$ -o $stdout\n'
        args += '#$ -e $stderr\n'
        if 'maxWallTime' in parameters :
            args += '#$ -l h_rt=$maxWallTime\n'
        if 'maxCpuTime' in parameters :
            args += '#$ -l cput=$maxCpuTime\n'
        if 'maxMemory' in parameters :
            args += '#$ -l mem_free=$maxMemoryM\n'
        if int(parameters['count']) > 1:
            args += '#$ -pe $parallel_env $count\n'
        args += '#$ -v %s\n' % (','.join(['%s=%s' %(k, v) for k, v in list(parameters['environment'].items())]))
        args += '\n'
        args += '$executable\n'
        return Template(args).safe_substitute(parameters)



#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

__all__ = ["communicators", "core", "managers", "utils", "commands", "api"]

from ._version import __version__
del _version

#__version__  = '2.6.15'
__author__   = 'Valvanuz Fernández-Quiruelas, Markel Garcia, Carlos Blanco, Antonio Minondo and Antonio S. Cofino (@cofinoa)'

import sys
import os
import logging.config
import logging
from os.path import dirname , join , expanduser , exists , abspath
from  shutil import copytree

if sys.version_info<(3,5):
    exit( 'The version number of Python has to be >= 3.5' )

##############################################
# Default values used in DRM4G ENV variables #
##############################################
HOME                 = expanduser(os.environ.get( 'HOME', '~' ))
DRM4G_DIR            = expanduser(os.environ.get( 'DRM4G_DIR'           , join( HOME     , '.drm4g' )))
DRM4G_DIR_VAR        = expanduser(os.environ.get( 'DRM4G_DIR_VAR'       , join( DRM4G_DIR, 'var' )))
DRM4G_DIR_ETC        = expanduser(os.environ.get( 'DRM4G_DIR_ETC'       , join( DRM4G_DIR, 'etc' )))
DRM4G_RESOURCES_CONF = expanduser(os.environ.get( 'DRM4G_RESOURCES_CONF', join( DRM4G_DIR_ETC, 'resources.conf' )))
DRM4G_LOGGER_CONF    = expanduser(os.environ.get( 'DRM4G_LOGGER_CONF'   , join( DRM4G_DIR_ETC, 'logger.conf')))
DRM4G_GWD_CONF       = expanduser(os.environ.get( 'DRM4G_GWD_CONF'      , join( DRM4G_DIR_ETC, 'gwd.conf')))
DRM4G_SCHED_CONF     = expanduser(os.environ.get( 'DRM4G_SCHED_CONF'    , join( DRM4G_DIR_ETC, 'sched.conf')))

#ENV variable shouldn't be used outside of GWD and its components. 
os.environ[ 'GW_LOCATION' ] = DRM4G_DIR

#Populate and prepare DRM4G_DIR
if exists( DRM4G_DIR ) is False  :
    logging.warning( "Creating a DRM4G local configuration in '%s'" %  DRM4G_DIR )
    acct_dir = join ( DRM4G_DIR_VAR, 'acct' )
    logging.warning( "Creating '%s' directory" % acct_dir )
    os.makedirs( acct_dir )
    src_dir  = join ( abspath( dirname( __file__ ) ), 'conf' )
    logging.warning( "Copying from '%s' to '%s'" % ( src_dir , DRM4G_DIR_ETC ) )
    copytree( src_dir , DRM4G_DIR_ETC )

##
# Configure logger
##
logging.config.fileConfig(DRM4G_LOGGER_CONF, {"DRM4G_DIR": DRM4G_DIR})
console_logger = logging.getLogger('console')

REMOTE_VOS_DIR  = join( DRM4G_DIR , 'security') #TODO: TO BE DEPRECATED

COMMUNICATORS = {
    "ssh"          : "drm4g.communicators.ssh",
    "ssh_fabric"   : "drm4g.communicators.ssh_fabric",
    "pk_ssh"       : "drm4g.communicators.ssh",
    "op_ssh"       : "drm4g.communicators.openssh",
    "local"        : "drm4g.communicators.local",
}
RESOURCE_MANAGERS = {
    "pbs"          : "drm4g.managers.pbs",
    "sge"          : "drm4g.managers.sge",
    "fork"         : "drm4g.managers.fork",
    "none"         : "drm4g.managers.fork",
    "lsf"          : "drm4g.managers.lsf",
    "loadleveler"  : "drm4g.managers.loadleveler",
    "cream"        : "drm4g.managers.cream",
    "slurm"        : "drm4g.managers.slurm",
    "mnslurm"      : "drm4g.managers.marenostrum",
    "slurm_res"    : "drm4g.managers.slurm_res",
    "neptuno"      : "drm4g.managers.neptuno",
    #"rocci"        : "drm4g.managers.rocci",
}





#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import re
import drm4g.managers
from string import Template


# The programs needed by these utilities. If they are not in a location
# accessible by PATH, specify their location here.
LLCLASS  = 'llclass'    #show class information
LLSUBMIT = 'llsubmit'   #submit ajob
LLQ      = 'llq'        #show jobs' status
LLCANCEL = 'llcancel'   #delete ajob

class Resource (drm4g.managers.Resource):
    pass

class Job (drm4g.managers.Job):

    #loadleveler job status <--> GridWay job status
    states_loadleveler = {'CA': 'DONE',
                  'CK': 'ACTIVE',
                  'C' : 'PENDING',
                  'CP': 'ACTIVE',
                  'D' : 'PENDING',
                  'I' : 'PENDING',
                  'NQ': 'SUSPENDED',
                  'NR': 'SUSPENDED',
                  'P' : 'ACTIVE',
                  'E' : 'ACTIVE',
                  'EP': 'ACTIVE',
                  'X' : 'DONE',
                  'XP': 'SUSPENDED',
                  'RM': 'DONE',
                  'RP': 'SUSPENDED',
                  'MP': 'ACTIVE',
                  'R' : 'ACTIVE',
                  'ST': 'ACTIVE',
                  'S' : 'PENDING',
                  'TX': 'PENDING',
                  'HS': 'PENDING',
                  'H' : 'PENDING',
                  'V' : 'DONE',
                  'VP': 'PENDING',
                }
    re_submit=re.compile(r"The job \"(\S+)\" has been submitted")

    def jobSubmit(self, pathScript):
        out, err = self.Communicator.execCommand('%s %s' % (LLSUBMIT, pathScript))
        job_id = self.re_submit.search(out).group(1)
        return job_id

    def jobStatus(self):
        command = LLQ + ' -f %st ' + self.JobId
        out, err = self.Communicator.execCommand(command)
        if "There is currently no job status to report" in out :
            return 'DONE'
        else:
            status = out.split('\n')[2].strip()
            return self.states_loadleveler.setdefault(status, 'UNKNOWN')

    def jobCancel(self):
        out, err = self.Communicator.execCommand('%s %s' % (LLCANCEL, self.JobId))
        if err:
            raise drm4g.managers.JobException(' '.join(err.split('\n')))

    def jobTemplate(self, parameters):
        args  = '#!/bin/bash\n'
        args += '#@ job_name = JID_%s\n' % (parameters['environment']['GW_JOB_ID'])
        if parameters['queue'] != 'default':
            args += '#@ class    = $queue\n'
        args += '#@ output   = $stdout\n'
        args += '#@ error    = $stderr\n'
        if int(parameters['count']) > 1 :
            args += '#@ job_type  = parallel\n'
        else:
            args += '#@ job_type  = serial\n'
        args += '#@ node = $count\n'
        if 'maxWallTime' in parameters:
            args += '#@ wall_clock_limit = $maxWallTime\n'
        if 'maxCpuTime' in parameters:
            args += '#@ job_cpu_limit = $maxCpuTime\n'
        if 'maxMemory' in parameters :
            args += '#@ resources = ConsumableMemory($maxMemory)\n'
        if 'ppn' in parameters:
            args += '#@ tasks_per_node = $ppn'
        if 'project' in parameters:
            args += '#@ account_no = $project\n'
        args += '#@ queue\n'
        args += ''.join(['export %s=%s\n' % (k, v) for k, v in list(parameters['environment'].items())])
        args += '\n'
        args += '$executable\n'
        return Template(args).safe_substitute(parameters)




#
# Copyright 2021 Santander Meteorology Group (UC-CSIC)
#
# Licensed under the EUPL, Version 1.1 only (the
# "Licence");
# You may not use this work except in compliance with the
# Licence.
# You may obtain a copy of the Licence at:
#
# http://ec.europa.eu/idabc/eupl
#
# Unless required by applicable law or agreed to in
# writing, software distributed under the Licence is
# distributed on an "AS IS" basis,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# See the Licence for the specific language governing
# permissions and limitations under the Licence.
#

import os
import re
import json
import pickle
import logging
import subprocess

def exec_cmd( cmd, stdin = subprocess.PIPE, stdout = subprocess.PIPE,
              stderr = subprocess.PIPE, env = os.environ ):
    """
    Execute shell commands.
    """
    logging.info( "Executing command ... " + cmd )
    if type( stdin ) == str :
        p      = subprocess.Popen( cmd, shell = True, stdout = stdout,
                                   stdin = subprocess.PIPE, stderr = stderr, env = env )
        p.stdin.write( stdin + "\n" )
        p.stdin.flush()
    else :
        p      = subprocess.Popen( cmd, shell = True, stdout = stdout,
                               stdin = stdin, stderr = stderr, env = env )
    output = p.stdout.read().strip() + p.stderr.read().strip()
    code   = p.wait()
    return code, output

def is_ip_private(ip):
    """
    Determinate if an IP is private
    """
    priv_lo = re.compile("^127\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
    priv_24 = re.compile("^10\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
    priv_20 = re.compile("^192\.168\.\d{1,3}.\d{1,3}$")
    priv_16 = re.compile("^172.(1[6-9]|2[0-9]|3[0-1]).[0-9]{1,3}.[0-9]{1,3}$")
    return True if priv_lo.match(ip) or priv_24.match(ip) or priv_20.match(ip) or priv_16.match(ip) else False

def download_proxy( myproxy_server, myproxy_password, vo, proxy_file, username ) :
    """
    Function to download proxies.
    """
    cmd = "MYPROXY_SERVER=%s myproxy-logon --voms %s -o %s --username %s -S" % ( myproxy_server, vo, proxy_file, username )
    return exec_cmd( cmd, stdin = myproxy_password )

def voms_proxy( proxy, vo, signed_proxy ) :
    """
    Function to sign proxies.
    """
    cmd = "X509_USER_PROXY=%s voms-proxy-init -ignorewarn " \
          "-timeout 30 -valid 24:00 -q -voms %s -noregen -out %s" % ( proxy, vo, signed_proxy )
    return exec_cmd( cmd )

def generate_key( file_name ) :
    """
    Function to generate an authentication key.
    """
    cmd = 'rm -rf %s; ssh-keygen -t rsa -b 2048 -f %s -N ""' % ( file_name, file_name )
    return exec_cmd( cmd )

def read_key( file_name ) :
    """
    Function to read a file key.
    """
    with open( file_name, "r" ) as file :
        pub = file.readline().rstrip()
    return pub

def save_pkl( obj_config, file_name ) :
    """
    Save a python object into a pickle file.
    """
    with open( file_name, "w" ) as f :
        pickle.dump( obj_config, f )

def load_pkl( file_name ) :
    """
    Load a python object back from the pickle file.
    """
    with open( file_name, "r" ) as f :
        return pickle.load( f )

def _decode_list(data):
    rv = []
    for item in data:
        if isinstance(item, unicode):
            item = item.encode('utf-8')
        elif isinstance(item, list):
            item = _decode_list(item)
        elif isinstance(item, dict):
            item = _decode_dict(item)
        rv.append(item)
    return rv

def _decode_dict(data):
    rv = {}
    for key, value in data.iteritems():
        if isinstance(key, unicode):
            key = key.encode('utf-8')
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        elif isinstance(value, list):
            value = _decode_list(value)
        elif isinstance(value, dict):
            value = _decode_dict(value)
        rv[key] = value
    return rv

def save_json( obj_config, file_name ) :
    """
    Save a python object into a json file.
    """
    with open( file_name, "w" ) as f :
        json.dump( obj_config, f )

def load_json( file_name ) :
    """
    Load a python object back from the json file.
    """
    with open( file_name, "r" ) as f :
        return json.load( f, object_hook=_decode_dict)

